from __future__ import print_function

from rpython.rtyper.annlowlevel import cast_instance_to_gcref
from rpython.rlib.debug import debug_print, debug_start, debug_stop
from rpython.jit.backend.llsupport.regalloc import FrameManager, \
        RegisterManager, TempVar, compute_vars_longevity, BaseRegalloc, \
        get_scale
from rpython.jit.backend.arm import registers as r
from rpython.jit.backend.arm import conditions as c
from rpython.jit.backend.arm import locations
from rpython.jit.backend.arm.locations import imm, get_fp_offset
from rpython.jit.backend.arm.helper.regalloc import (
                                                    prepare_unary_cmp,
                                                    prepare_op_ri,
                                                    prepare_int_cmp,
                                                    prepare_unary_op,
                                                    prepare_two_regs_op,
                                                    prepare_float_cmp,
                                                    check_imm_arg,
                                                    check_imm_box,
                                                    VMEM_imm_size,
                                                    default_imm_size,
                                                    )
from rpython.jit.backend.arm.jump import remap_frame_layout_mixed
from rpython.jit.backend.arm.arch import WORD, JITFRAME_FIXED_SIZE
from rpython.jit.codewriter import longlong
from rpython.jit.metainterp.history import (Const, ConstInt, ConstFloat,
                                            ConstPtr,
                                            INT, REF, FLOAT)
from rpython.jit.metainterp.history import TargetToken
from rpython.jit.metainterp.resoperation import rop
from rpython.jit.backend.llsupport.descr import ArrayDescr
from rpython.jit.backend.llsupport.gcmap import allocate_gcmap
from rpython.jit.backend.llsupport import symbolic
from rpython.rtyper.lltypesystem import lltype, rffi, rstr, llmemory
from rpython.rtyper.lltypesystem.lloperation import llop
from rpython.jit.codewriter.effectinfo import EffectInfo
from rpython.rlib.rarithmetic import r_uint
from rpython.jit.backend.llsupport.descr import CallDescr


class TempInt(TempVar):
    type = INT

    def __repr__(self):
        return "<TempInt at %s>" % (id(self),)


class TempPtr(TempVar):
    type = REF

    def __repr__(self):
        return "<TempPtr at %s>" % (id(self),)


class TempFloat(TempVar):
    type = FLOAT

    def __repr__(self):
        return "<TempFloat at %s>" % (id(self),)



class ARMFrameManager(FrameManager):

    def __init__(self, base_ofs):
        FrameManager.__init__(self)
        self.base_ofs = base_ofs

    def frame_pos(self, i, box_type):
        return locations.StackLocation(i, get_fp_offset(self.base_ofs, i), box_type)

    @staticmethod
    def frame_size(type):
        if type == FLOAT:
            return  2
        return 1

    @staticmethod
    def get_loc_index(loc):
        assert loc.is_stack()
        return loc.position


def void(self, op, fcond):
    return []

class ARMRegisterManager(RegisterManager):
    FORBID_TEMP_BOXES = True

    def return_constant(self, v, forbidden_vars=[], selected_reg=None):
        self._check_type(v)
        if isinstance(v, Const):
            if isinstance(v, ConstPtr):
                tp = REF
            elif isinstance(v, ConstFloat):
                tp = FLOAT
            else:
                tp = INT
            loc = self.get_scratch_reg(tp,
                    forbidden_vars,
                    selected_reg=selected_reg)
            immvalue = self.convert_to_imm(v)
            self.assembler.load(loc, immvalue)
            return loc
        else:
            return RegisterManager.return_constant(self, v,
                                    forbidden_vars, selected_reg)


class VFPRegisterManager(ARMRegisterManager):
    all_regs = r.all_vfp_regs
    box_types = [FLOAT]
    save_around_call_regs = r.all_vfp_regs

    def convert_to_imm(self, c):
        adr = self.assembler.datablockwrapper.malloc_aligned(8, 8)
        x = c.getfloatstorage()
        rffi.cast(rffi.CArrayPtr(longlong.FLOATSTORAGE), adr)[0] = x
        return locations.ConstFloatLoc(adr)

    def __init__(self, longevity, frame_manager=None, assembler=None):
        RegisterManager.__init__(self, longevity, frame_manager, assembler)

    def after_call(self, v):
        """ Adjust registers according to the result of the call,
        which is in variable v.
        """
        self._check_type(v)
        reg = self.force_allocate_reg(v, selected_reg=r.d0)
        return reg

    def get_scratch_reg(self, type=FLOAT, forbidden_vars=[], selected_reg=None):
        assert type == FLOAT  # for now
        box = TempFloat()
        reg = self.force_allocate_reg(box, forbidden_vars=forbidden_vars,
                                                    selected_reg=selected_reg)
        self.temp_boxes.append(box)
        return reg


class CoreRegisterManager(ARMRegisterManager):
    all_regs = r.all_regs
    box_types = None       # or a list of acceptable types
    no_lower_byte_regs = all_regs
    save_around_call_regs = r.caller_resp
    frame_reg = r.fp

    def __init__(self, longevity, frame_manager=None, assembler=None):
        RegisterManager.__init__(self, longevity, frame_manager, assembler)

    def call_result_location(self, v):
        return r.r0

    def convert_to_imm(self, c):
        if isinstance(c, ConstInt):
            val = rffi.cast(rffi.INT, c.value)
            return locations.ImmLocation(val)
        else:
            assert isinstance(c, ConstPtr)
            return locations.ImmLocation(rffi.cast(lltype.Signed, c.value))
        assert 0

    def get_scratch_reg(self, type=INT, forbidden_vars=[], selected_reg=None):
        assert type == INT or type == REF
        box = None
        if type == INT:
            box = TempInt()
        else:
            box = TempPtr()
        reg = self.force_allocate_reg(box, forbidden_vars=forbidden_vars,
                                                    selected_reg=selected_reg)
        self.temp_boxes.append(box)
        return reg

    def get_free_reg(self):
        free_regs = self.free_regs
        for i in range(len(free_regs) - 1, -1, -1):
            if free_regs[i] in self.save_around_call_regs:
                continue
            return free_regs[i]


class Regalloc(BaseRegalloc):

    def __init__(self, assembler):
        self.cpu = assembler.cpu
        self.assembler = assembler
        self.frame_manager = None
        self.jump_target_descr = None
        self.final_jump_op = None

    def loc(self, var):
        if var.type == FLOAT:
            return self.vfprm.loc(var)
        else:
            return self.rm.loc(var)

    def position(self):
        return self.rm.position

    def next_instruction(self):
        self.rm.next_instruction()
        self.vfprm.next_instruction()

    def _check_invariants(self):
        self.rm._check_invariants()
        self.vfprm._check_invariants()

    def stays_alive(self, v):
        if v.type == FLOAT:
            return self.vfprm.stays_alive(v)
        else:
            return self.rm.stays_alive(v)

    def call_result_location(self, v):
        if v.type == FLOAT:
            return self.vfprm.call_result_location(v)
        else:
            return self.rm.call_result_location(v)

    def after_call(self, v):
        if v.type == 'v':
            return
        if v.type == FLOAT:
            return self.vfprm.after_call(v)
        else:
            return self.rm.after_call(v)

    def force_allocate_reg(self, var, forbidden_vars=[], selected_reg=None,
                           need_lower_byte=False):
        if var.type == FLOAT:
            return self.vfprm.force_allocate_reg(var, forbidden_vars,
                                               selected_reg, need_lower_byte)
        else:
            return self.rm.force_allocate_reg(var, forbidden_vars,
                                              selected_reg, need_lower_byte)

    def force_allocate_reg_or_cc(self, var, forbidden_vars=[]):
        assert var.type == INT
        if self.next_op_can_accept_cc(self.operations, self.rm.position):
            # hack: return the 'fp' location to mean "lives in CC".  This
            # fp will not actually be used, and the location will be freed
            # after the next op as usual.
            self.rm.force_allocate_frame_reg(var)
            return r.fp
        else:
            # else, return a regular register (not fp).
            return self.rm.force_allocate_reg(var)

    def try_allocate_reg(self, v, selected_reg=None, need_lower_byte=False):
        if v.type == FLOAT:
            return self.vfprm.try_allocate_reg(v, selected_reg,
                                                            need_lower_byte)
        else:
            return self.rm.try_allocate_reg(v, selected_reg, need_lower_byte)

    def possibly_free_var(self, var):
        if var.type == FLOAT:
            self.vfprm.possibly_free_var(var)
        else:
            self.rm.possibly_free_var(var)

    def possibly_free_vars_for_op(self, op):
        for i in range(op.numargs()):
            var = op.getarg(i)
            if var is not None:  # xxx kludgy
                self.possibly_free_var(var)

    def possibly_free_vars(self, vars):
        for var in vars:
            if var is not None:  # xxx kludgy
                self.possibly_free_var(var)

    def get_scratch_reg(self, type, forbidden_vars=[], selected_reg=None):
        if type == FLOAT:
            return self.vfprm.get_scratch_reg(type, forbidden_vars,
                                                                selected_reg)
        else:
            return self.rm.get_scratch_reg(type, forbidden_vars, selected_reg)

    def get_free_reg(self):
        return self.rm.get_free_reg()

    def free_temp_vars(self):
        self.rm.free_temp_vars()
        self.vfprm.free_temp_vars()

    def make_sure_var_in_reg(self, var, forbidden_vars=[],
                         selected_reg=None, need_lower_byte=False):
        if var.type == FLOAT:
            return self.vfprm.make_sure_var_in_reg(var, forbidden_vars,
                                        selected_reg, need_lower_byte)
        else:
            return self.rm.make_sure_var_in_reg(var, forbidden_vars,
                                        selected_reg, need_lower_byte)

    def convert_to_imm(self, value):
        if isinstance(value, ConstInt):
            return self.rm.convert_to_imm(value)
        else:
            assert isinstance(value, ConstFloat)
            return self.vfprm.convert_to_imm(value)

    def _prepare(self, inputargs, operations, allgcrefs):
        cpu = self.cpu
        self.fm = ARMFrameManager(cpu.get_baseofs_of_frame_field())
        self.frame_manager = self.fm
        operations = cpu.gc_ll_descr.rewrite_assembler(cpu, operations,
                                                       allgcrefs)
        # compute longevity of variables
        longevity = compute_vars_longevity(inputargs, operations)
        self.longevity = longevity
        fm = self.frame_manager
        asm = self.assembler
        self.vfprm = VFPRegisterManager(longevity, fm, asm)
        self.rm = CoreRegisterManager(longevity, fm, asm)
        return operations

    def prepare_loop(self, inputargs, operations, looptoken, allgcrefs):
        operations = self._prepare(inputargs, operations, allgcrefs)
        self._set_initial_bindings(inputargs, looptoken)
        self.possibly_free_vars(list(inputargs))
        return operations

    def prepare_bridge(self, inputargs, arglocs, operations, allgcrefs,
                       frame_info):
        operations = self._prepare(inputargs, operations, allgcrefs)
        self._update_bindings(arglocs, inputargs)
        return operations

    def get_final_frame_depth(self):
        return self.frame_manager.get_frame_depth()

    def _update_bindings(self, locs, inputargs):
        used = {}
        i = 0
        for loc in locs:
            if loc is None:
                loc = r.fp
            arg = inputargs[i]
            i += 1
            if loc.is_core_reg():
                self.rm.reg_bindings[arg] = loc
                used[loc] = None
            elif loc.is_vfp_reg():
                self.vfprm.reg_bindings[arg] = loc
                used[loc] = None
            else:
                assert loc.is_stack()
                self.frame_manager.bind(arg, loc)

        # XXX combine with x86 code and move to llsupport
        self.rm.free_regs = []
        for reg in self.rm.all_regs:
            if reg not in used:
                self.rm.free_regs.append(reg)
        self.vfprm.free_regs = []
        for reg in self.vfprm.all_regs:
            if reg not in used:
                self.vfprm.free_regs.append(reg)
        # note: we need to make a copy of inputargs because possibly_free_vars
        # is also used on op args, which is a non-resizable list
        self.possibly_free_vars(list(inputargs))
        self.fm.finish_binding()
        self._check_invariants()

    def get_gcmap(self, forbidden_regs=[], noregs=False):
        frame_depth = self.fm.get_frame_depth()
        gcmap = allocate_gcmap(self.assembler,
                        frame_depth, JITFRAME_FIXED_SIZE)
        for box, loc in self.rm.reg_bindings.iteritems():
            if loc in forbidden_regs:
                continue
            if box.type == REF and self.rm.is_still_alive(box):
                assert not noregs
                assert loc.is_core_reg()
                #val = self.cpu.all_reg_indexes[loc.value]
                # ^^^ That is the correct way to write it down, but as a
                #     special case in the arm backend only, this is equivalent
                #     to just the line below:
                val = loc.value
                gcmap[val // WORD // 8] |= r_uint(1) << (val % (WORD * 8))
        for box, loc in self.fm.bindings_iteritems():
            if box.type == REF and self.rm.is_still_alive(box):
                assert loc.is_stack()
                val = loc.position + JITFRAME_FIXED_SIZE
                gcmap[val // WORD // 8] |= r_uint(1) << (val % (WORD * 8))
        return gcmap

    # ------------------------------------------------------------
    def perform_enter_portal_frame(self, op):
        self.assembler.enter_portal_frame(op)

    def perform_leave_portal_frame(self, op):
        self.assembler.leave_portal_frame(op)

    def perform_extra(self, op, args, fcond):
        return self.assembler.regalloc_emit_extra(op, args, fcond, self)

    def force_spill_var(self, var):
        if var.type == FLOAT:
            self.vfprm.force_spill_var(var)
        else:
            self.rm.force_spill_var(var)

    def before_call(self, save_all_regs=False):
        self.rm.before_call(save_all_regs=save_all_regs)
        self.vfprm.before_call(save_all_regs=save_all_regs)

    def _sync_var(self, v):
        if v.type == FLOAT:
            self.vfprm._sync_var(v)
        else:
            self.rm._sync_var(v)

    def _prepare_op_int_add(self, op, fcond):
        boxes = op.getarglist()
        a0, a1 = boxes
        imm_a0 = check_imm_box(a0)
        imm_a1 = check_imm_box(a1)
        if not imm_a0 and imm_a1:
            l0 = self.make_sure_var_in_reg(a0, boxes)
            l1 = self.convert_to_imm(a1)
        elif imm_a0 and not imm_a1:
            l0 = self.convert_to_imm(a0)
            l1 = self.make_sure_var_in_reg(a1, boxes)
        else:
            l0 = self.make_sure_var_in_reg(a0, boxes)
            l1 = self.make_sure_var_in_reg(a1, boxes)
        return [l0, l1]

    def prepare_op_int_add(self, op, fcond):
        locs = self._prepare_op_int_add(op, fcond)
        self.possibly_free_vars_for_op(op)
        self.free_temp_vars()
        res = self.force_allocate_reg(op)
        return locs + [res]

    prepare_op_nursery_ptr_increment = prepare_op_int_add

    def _prepare_op_int_sub(self, op, fcond):
        a0, a1 = boxes = op.getarglist()
        imm_a0 = check_imm_box(a0)
        imm_a1 = check_imm_box(a1)
        if not imm_a0 and imm_a1:
            l0 = self.make_sure_var_in_reg(a0, boxes)
            l1 = self.convert_to_imm(a1)
        elif imm_a0 and not imm_a1:
            l0 = self.convert_to_imm(a0)
            l1 = self.make_sure_var_in_reg(a1, boxes)
        else:
            l0 = self.make_sure_var_in_reg(a0, boxes)
            l1 = self.make_sure_var_in_reg(a1, boxes)
        return [l0, l1]

    def prepare_op_int_sub(self, op, fcond):
        locs = self._prepare_op_int_sub(op, fcond)
        self.possibly_free_vars_for_op(op)
        self.free_temp_vars()
        res = self.force_allocate_reg(op)
        return locs + [res]

    def prepare_op_int_mul(self, op, fcond):
        boxes = op.getarglist()
        a0, a1 = boxes

        reg1 = self.make_sure_var_in_reg(a0, forbidden_vars=boxes)
        reg2 = self.make_sure_var_in_reg(a1, forbidden_vars=boxes)

        self.possibly_free_vars(boxes)
        self.possibly_free_vars_for_op(op)
        res = self.force_allocate_reg(op)
        self.possibly_free_var(op)
        return [reg1, reg2, res]

    prepare_op_uint_mul_high = prepare_op_int_mul

    def prepare_op_int_force_ge_zero(self, op, fcond):
        argloc = self.make_sure_var_in_reg(op.getarg(0))
        resloc = self.force_allocate_reg(op, [op.getarg(0)])
        return [argloc, resloc]

    def prepare_op_int_signext(self, op, fcond):
        argloc = self.make_sure_var_in_reg(op.getarg(0))
        numbytes = op.getarg(1).getint()
        resloc = self.force_allocate_reg(op)
        return [argloc, imm(numbytes), resloc]

    prepare_op_int_and = prepare_op_ri('int_and')
    prepare_op_int_or = prepare_op_ri('int_or')
    prepare_op_int_xor = prepare_op_ri('int_xor')
    prepare_op_int_lshift = prepare_op_ri('int_lshift', imm_size=0x1F,
                                        allow_zero=False, commutative=False)
    prepare_op_int_rshift = prepare_op_ri('int_rshift', imm_size=0x1F,
                                        allow_zero=False, commutative=False)
    prepare_op_uint_rshift = prepare_op_ri('uint_rshift', imm_size=0x1F,
                                        allow_zero=False, commutative=False)

    prepare_op_int_lt = prepare_int_cmp
    prepare_op_int_le = prepare_int_cmp
    prepare_op_int_eq = prepare_int_cmp
    prepare_op_int_ne = prepare_int_cmp
    prepare_op_int_gt = prepare_int_cmp
    prepare_op_int_ge = prepare_int_cmp

    prepare_op_uint_le = prepare_int_cmp
    prepare_op_uint_gt = prepare_int_cmp

    prepare_op_uint_lt = prepare_int_cmp
    prepare_op_uint_ge = prepare_int_cmp

    prepare_op_ptr_eq = prepare_op_instance_ptr_eq = prepare_op_int_eq
    prepare_op_ptr_ne = prepare_op_instance_ptr_ne = prepare_op_int_ne

    prepare_op_int_add_ovf = prepare_op_int_add
    prepare_op_int_sub_ovf = prepare_op_int_sub
    prepare_op_int_mul_ovf = prepare_op_int_mul

    prepare_op_int_is_true = prepare_unary_cmp
    prepare_op_int_is_zero = prepare_unary_cmp

    prepare_op_int_neg = prepare_unary_op
    prepare_op_int_invert = prepare_unary_op

    def _prepare_op_call(self, op, fcond):
        calldescr = op.getdescr()
        assert calldescr is not None
        effectinfo = calldescr.get_extra_info()
        if effectinfo is not None:
            oopspecindex = effectinfo.oopspecindex
            if oopspecindex in (EffectInfo.OS_LLONG_ADD,
                            EffectInfo.OS_LLONG_SUB,
                            EffectInfo.OS_LLONG_AND,
                            EffectInfo.OS_LLONG_OR,
                            EffectInfo.OS_LLONG_XOR):
                if self.cpu.cpuinfo.neon:
                    args = self._prepare_llong_binop_xx(op, fcond)
                    self.perform_extra(op, args, fcond)
                    return
            elif oopspecindex == EffectInfo.OS_LLONG_TO_INT:
                args = self._prepare_llong_to_int(op, fcond)
                self.perform_extra(op, args, fcond)
                return
            elif oopspecindex == EffectInfo.OS_MATH_SQRT:
                args = self._prepare_op_math_sqrt(op, fcond)
                self.perform_extra(op, args, fcond)
                return
            elif oopspecindex == EffectInfo.OS_THREADLOCALREF_GET:
                args = self._prepare_threadlocalref_get(op, fcond)
                self.perform_extra(op, args, fcond)
                return
            #elif oopspecindex == EffectInfo.OS_MATH_READ_TIMESTAMP:
            #    ...
        return self._prepare_call(op)

    prepare_op_call_i = _prepare_op_call
    prepare_op_call_r = _prepare_op_call
    prepare_op_call_f = _prepare_op_call
    prepare_op_call_n = _prepare_op_call

    def _prepare_call(self, op, save_all_regs=False, first_arg_index=1):
        args = [None] * (op.numargs() + 3)
        calldescr = op.getdescr()
        assert isinstance(calldescr, CallDescr)
        assert len(calldescr.arg_classes) == op.numargs() - first_arg_index

        for i in range(op.numargs()):
            args[i + 3] = self.loc(op.getarg(i))

        size = calldescr.get_result_size()
        sign = calldescr.is_result_signed()
        if sign:
            sign_loc = imm(1)
        else:
            sign_loc = imm(0)
        args[1] = imm(size)
        args[2] = sign_loc

        effectinfo = calldescr.get_extra_info()
        if save_all_regs:
            gc_level = 2
        elif effectinfo is None or effectinfo.check_can_collect():
            gc_level = 1
        else:
            gc_level = 0

        args[0] = self._call(op, args, gc_level)
        return args

    def _call(self, op, arglocs, gc_level):
        # spill variables that need to be saved around calls:
        # gc_level == 0: callee cannot invoke the GC
        # gc_level == 1: can invoke GC, save all regs that contain pointers
        # gc_level == 2: can force, save all regs
        save_all_regs = gc_level == 2
        self.vfprm.before_call(save_all_regs=save_all_regs)
        if gc_level == 1 and self.cpu.gc_ll_descr.gcrootmap:
            save_all_regs = 2
        self.rm.before_call(save_all_regs=save_all_regs)
        resloc = self.after_call(op)
        return resloc

    def prepare_op_check_memory_error(self, op, fcond):
        argloc = self.make_sure_var_in_reg(op.getarg(0))
        return [argloc]

    def _prepare_llong_binop_xx(self, op, fcond):
        # arg 0 is the address of the function
        loc0 = self.make_sure_var_in_reg(op.getarg(1))
        loc1 = self.make_sure_var_in_reg(op.getarg(2))
        self.possibly_free_vars_for_op(op)
        self.free_temp_vars()
        res = self.vfprm.force_allocate_reg(op)
        return [loc0, loc1, res]

    def _prepare_llong_to_int(self, op, fcond):
        loc0 = self.make_sure_var_in_reg(op.getarg(1))
        res = self.force_allocate_reg(op)
        return [loc0, res]

    def _prepare_threadlocalref_get(self, op, fcond):
        ofs_loc = imm(op.getarg(1).getint())
        calldescr = op.getdescr()
        size_loc = imm(calldescr.get_result_size())
        sign_loc = imm(calldescr.is_result_signed())
        res_loc = self.force_allocate_reg(op)
        return [ofs_loc, size_loc, sign_loc, res_loc]

    def _prepare_guard(self, op, args=None):
        if args is None:
            args = []
        args.append(imm(self.frame_manager.get_frame_depth()))
        for arg in op.getfailargs():
            if arg:
                args.append(self.loc(arg))
            else:
                args.append(None)
        return args

    def prepare_op_finish(self, op, fcond):
        # the frame is in fp, but we have to point where in the frame is
        # the potential argument to FINISH
        if op.numargs() == 1:
            loc = self.make_sure_var_in_reg(op.getarg(0))
            locs = [loc]
        else:
            locs = []
        return locs

    def load_condition_into_cc(self, box):
        if self.assembler.guard_success_cc == c.cond_none:
            loc = self.loc(box)
            if not loc.is_core_reg():
                assert loc.is_stack()
                self.assembler.regalloc_mov(loc, r.lr)
                loc = r.lr
            self.assembler.mc.CMP_ri(loc.value, 0)
            self.assembler.guard_success_cc = c.NE

    def _prepare_guard_cc(self, op, fcond):
        self.load_condition_into_cc(op.getarg(0))
        args = self._prepare_guard(op, [])
        return args

    prepare_op_guard_true = _prepare_guard_cc
    prepare_op_guard_false = _prepare_guard_cc
    prepare_op_guard_nonnull = _prepare_guard_cc
    prepare_op_guard_isnull = _prepare_guard_cc

    def prepare_op_guard_value(self, op, fcond):
        boxes = op.getarglist()
        a0, a1 = boxes
        imm_a1 = check_imm_box(a1)
        l0 = self.make_sure_var_in_reg(a0, boxes)
        op.getdescr().make_a_counter_per_value(op,
            self.cpu.all_reg_indexes[l0.value])
        if not imm_a1:
            l1 = self.make_sure_var_in_reg(a1, boxes)
        else:
            l1 = self.convert_to_imm(a1)
        arglocs = self._prepare_guard(op, [l0, l1])
        self.possibly_free_vars(op.getarglist())
        self.possibly_free_vars(op.getfailargs())
        return arglocs

    def prepare_op_guard_no_overflow(self, op, fcond):
        locs = self._prepare_guard(op)
        self.possibly_free_vars(op.getfailargs())
        return locs

    prepare_op_guard_overflow = prepare_op_guard_no_overflow
    prepare_op_guard_not_invalidated = prepare_op_guard_no_overflow
    prepare_op_guard_not_forced = prepare_op_guard_no_overflow

    def prepare_op_guard_exception(self, op, fcond):
        boxes = op.getarglist()
        arg0 = ConstInt(rffi.cast(lltype.Signed, op.getarg(0).getint()))
        loc = self.make_sure_var_in_reg(arg0)
        loc1 = self.get_scratch_reg(INT, boxes)
        if op in self.longevity:
            resloc = self.force_allocate_reg(op, boxes)
            self.possibly_free_var(op)
        else:
            resloc = None
        pos_exc_value = imm(self.cpu.pos_exc_value())
        pos_exception = imm(self.cpu.pos_exception())
        arglocs = self._prepare_guard(op,
                    [loc, loc1, resloc, pos_exc_value, pos_exception])
        return arglocs

    def prepare_op_save_exception(self, op, fcond):
        resloc = self.force_allocate_reg(op)
        return [resloc]
    prepare_op_save_exc_class = prepare_op_save_exception

    def prepare_op_restore_exception(self, op, fcond):
        boxes = op.getarglist()
        loc0 = self.make_sure_var_in_reg(op.getarg(0), boxes)  # exc class
        loc1 = self.make_sure_var_in_reg(op.getarg(1), boxes)  # exc instance
        return [loc0, loc1]

    def prepare_op_guard_no_exception(self, op, fcond):
        loc = self.make_sure_var_in_reg(ConstInt(self.cpu.pos_exception()))
        arglocs = self._prepare_guard(op, [loc])
        return arglocs

    def prepare_op_guard_class(self, op, fcond):
        assert not isinstance(op.getarg(0), Const)
        boxes = op.getarglist()

        x = self.make_sure_var_in_reg(boxes[0], boxes)
        y_val = rffi.cast(lltype.Signed, boxes[1].getint())
        return self._prepare_guard(op, [x, imm(y_val)])

    prepare_op_guard_nonnull_class = prepare_op_guard_class
    prepare_op_guard_gc_type = prepare_op_guard_class
    prepare_op_guard_subclass = prepare_op_guard_class

    def prepare_op_guard_is_object(self, op, fcond):
        loc_object = self.make_sure_var_in_reg(op.getarg(0))
        return self._prepare_guard(op, [loc_object])

    def compute_hint_frame_locations(self, operations):
        # optimization only: fill in the 'hint_frame_locations' dictionary
        # of rm and xrm based on the JUMP at the end of the loop, by looking
        # at where we would like the boxes to be after the jump.
        op = operations[-1]
        if op.getopnum() != rop.JUMP:
            return
        self.final_jump_op = op
        descr = op.getdescr()
        assert isinstance(descr, TargetToken)
        if descr._ll_loop_code != 0:
            # if the target LABEL was already compiled, i.e. if it belongs
            # to some already-compiled piece of code
            self._compute_hint_frame_locations_from_descr(descr)
        #else:
        #   The loop ends in a JUMP going back to a LABEL in the same loop.
        #   We cannot fill 'hint_frame_locations' immediately, but we can
        #   wait until the corresponding prepare_op_label() to know where the
        #   we would like the boxes to be after the jump.

    def _compute_hint_frame_locations_from_descr(self, descr):
        arglocs = self.assembler.target_arglocs(descr)
        jump_op = self.final_jump_op
        assert len(arglocs) == jump_op.numargs()
        for i in range(jump_op.numargs()):
            box = jump_op.getarg(i)
            if not isinstance(box, Const):
                loc = arglocs[i]
                if loc is not None and loc.is_stack():
                    self.frame_manager.add_frame_pos_hint(box, loc)

    def prepare_op_jump(self, op, fcond):
        assert self.jump_target_descr is None
        descr = op.getdescr()
        assert isinstance(descr, TargetToken)
        self.jump_target_descr = descr
        arglocs = self.assembler.target_arglocs(descr)

        # get temporary locs
        tmploc = r.ip
        vfptmploc = r.vfp_ip

        # Part about non-floats
        src_locations1 = []
        dst_locations1 = []
        # Part about floats
        src_locations2 = []
        dst_locations2 = []

        # Build the four lists
        for i in range(op.numargs()):
            box = op.getarg(i)
            src_loc = self.loc(box)
            dst_loc = arglocs[i]
            if box.type != FLOAT:
                src_locations1.append(src_loc)
                dst_locations1.append(dst_loc)
            else:
                src_locations2.append(src_loc)
                dst_locations2.append(dst_loc)
        self.assembler.check_frame_before_jump(self.jump_target_descr)
        remap_frame_layout_mixed(self.assembler,
                                 src_locations1, dst_locations1, tmploc,
                                 src_locations2, dst_locations2, vfptmploc)
        return []

    def prepare_op_gc_store(self, op, fcond):
        boxes = op.getarglist()
        base_loc = self.make_sure_var_in_reg(boxes[0], boxes)
        ofs = boxes[1].getint()
        value_loc = self.make_sure_var_in_reg(boxes[2], boxes)
        size = boxes[3].getint()
        ofs_size = default_imm_size if size < 8 else VMEM_imm_size
        if check_imm_arg(ofs, size=ofs_size):
            ofs_loc = imm(ofs)
        else:
            ofs_loc = self.get_scratch_reg(INT, boxes)
            self.assembler.load(ofs_loc, imm(ofs))
        return [value_loc, base_loc, ofs_loc, imm(size)]

    def _prepare_op_gc_load(self, op, fcond):
        a0 = op.getarg(0)
        ofs = op.getarg(1).getint()
        nsize = op.getarg(2).getint()    # negative for "signed"
        base_loc = self.make_sure_var_in_reg(a0)
        immofs = imm(ofs)
        ofs_size = default_imm_size if abs(nsize) < 8 else VMEM_imm_size
        if check_imm_arg(ofs, size=ofs_size):
            ofs_loc = immofs
        else:
            ofs_loc = self.get_scratch_reg(INT, [a0])
            self.assembler.load(ofs_loc, immofs)
        self.possibly_free_vars_for_op(op)
        self.free_temp_vars()
        res_loc = self.force_allocate_reg(op)
        return [base_loc, ofs_loc, res_loc, imm(nsize)]

    prepare_op_gc_load_i = _prepare_op_gc_load
    prepare_op_gc_load_r = _prepare_op_gc_load
    prepare_op_gc_load_f = _prepare_op_gc_load

    def prepare_op_increment_debug_counter(self, op, fcond):
        boxes = op.getarglist()
        a0, = boxes
        base_loc = self.make_sure_var_in_reg(a0, boxes)
        value_loc = self.get_scratch_reg(INT, boxes)
        self.free_temp_vars()
        return [base_loc, value_loc]

    def prepare_op_gc_store_indexed(self, op, fcond):
        boxes = op.getarglist()
        base_loc = self.make_sure_var_in_reg(boxes[0], boxes)
        value_loc = self.make_sure_var_in_reg(boxes[2], boxes)
        index_loc = self.make_sure_var_in_reg(boxes[1], boxes)
        assert boxes[3].getint() == 1    # scale
        ofs = boxes[4].getint()
        size = boxes[5].getint()
        return [value_loc, base_loc, index_loc, imm(size), imm(ofs)]

    def _prepare_op_gc_load_indexed(self, op, fcond):
        boxes = op.getarglist()
        base_loc = self.make_sure_var_in_reg(boxes[0], boxes)
        index_loc = self.make_sure_var_in_reg(boxes[1], boxes)
        assert boxes[2].getint() == 1    # scale
        ofs = boxes[3].getint()
        nsize = boxes[4].getint()
        self.possibly_free_vars_for_op(op)
        self.free_temp_vars()
        res_loc = self.force_allocate_reg(op)
        return [res_loc, base_loc, index_loc, imm(nsize), imm(ofs)]

    prepare_op_gc_load_indexed_i = _prepare_op_gc_load_indexed
    prepare_op_gc_load_indexed_r = _prepare_op_gc_load_indexed
    prepare_op_gc_load_indexed_f = _prepare_op_gc_load_indexed

    prepare_op_zero_array = void

    def _prepare_op_same_as(self, op, fcond):
        arg = op.getarg(0)
        imm_arg = check_imm_box(arg)
        if imm_arg:
            argloc = self.convert_to_imm(arg)
        else:
            argloc = self.make_sure_var_in_reg(arg)
        self.possibly_free_vars_for_op(op)
        self.free_temp_vars()
        resloc = self.force_allocate_reg(op)
        return [argloc, resloc]

    prepare_op_cast_ptr_to_int = _prepare_op_same_as
    prepare_op_cast_int_to_ptr = _prepare_op_same_as
    prepare_op_same_as_i = _prepare_op_same_as
    prepare_op_same_as_r = _prepare_op_same_as
    prepare_op_same_as_f = _prepare_op_same_as

    def prepare_op_load_from_gc_table(self, op, fcond):
        resloc = self.force_allocate_reg(op)
        return [resloc]

    def prepare_op_load_effective_address(self, op, fcond):
        args = op.getarglist()
        arg0 = self.make_sure_var_in_reg(args[0], args)
        arg1 = self.make_sure_var_in_reg(args[1], args)
        res = self.force_allocate_reg(op)
        return [arg0, arg1, res]

    def prepare_op_call_malloc_nursery(self, op, fcond):
        size_box = op.getarg(0)
        assert isinstance(size_box, ConstInt)
        size = size_box.getint()
        # hint: try to move unrelated registers away from r0 and r1 now
        self.rm.spill_or_move_registers_before_call([r.r0, r.r1])

        self.rm.force_allocate_reg(op, selected_reg=r.r0)
        t = TempInt()
        self.rm.force_allocate_reg(t, selected_reg=r.r1)

        sizeloc = size_box.getint()
        gc_ll_descr = self.cpu.gc_ll_descr
        gcmap = self.get_gcmap([r.r0, r.r1])
        self.possibly_free_var(t)
        self.assembler.malloc_cond(
            gc_ll_descr.get_nursery_free_addr(),
            gc_ll_descr.get_nursery_top_addr(),
            sizeloc,
            gcmap
            )
        self.assembler._alignment_check()

    def prepare_op_call_malloc_nursery_varsize_frame(self, op, fcond):
        size_box = op.getarg(0)
        assert not isinstance(size_box, ConstInt) # we cannot have a const here!
        # sizeloc must be in a register, but we can free it now
        # (we take care explicitly of conflicts with r0 or r1)
        sizeloc = self.rm.make_sure_var_in_reg(size_box)
        self.rm.spill_or_move_registers_before_call([r.r0, r.r1]) # sizeloc safe
        self.rm.possibly_free_var(size_box)
        #
        self.rm.force_allocate_reg(op, selected_reg=r.r0)
        #
        t = TempInt()
        self.rm.force_allocate_reg(t, selected_reg=r.r1)
        #
        gcmap = self.get_gcmap([r.r0, r.r1])
        self.possibly_free_var(t)
        #
        gc_ll_descr = self.cpu.gc_ll_descr
        self.assembler.malloc_cond_varsize_frame(
            gc_ll_descr.get_nursery_free_addr(),
            gc_ll_descr.get_nursery_top_addr(),
            sizeloc,
            gcmap
            )
        self.assembler._alignment_check()

    def prepare_op_call_malloc_nursery_varsize(self, op, fcond):
        gc_ll_descr = self.cpu.gc_ll_descr
        if not hasattr(gc_ll_descr, 'max_size_of_young_obj'):
            raise Exception("unreachable code")
            # for boehm, this function should never be called
        arraydescr = op.getdescr()
        length_box = op.getarg(2)
        assert not isinstance(length_box, Const) # we cannot have a const here!
        # can only use spill_or_move_registers_before_call() as a hint if
        # we are sure that length_box stays alive and won't be freed now
        # (it should always be the case, see below, but better safe than sorry)
        if self.rm.stays_alive(length_box):
            self.rm.spill_or_move_registers_before_call([r.r0, r.r1])
        # the result will be in r0
        self.rm.force_allocate_reg(op, selected_reg=r.r0)
        # we need r1 as a temporary
        tmp_box = TempVar()
        self.rm.force_allocate_reg(tmp_box, selected_reg=r.r1)
        gcmap = self.get_gcmap([r.r0, r.r1]) # allocate the gcmap *before*
        self.rm.possibly_free_var(tmp_box)
        # length_box always survives: it's typically also present in the
        # next operation that will copy it inside the new array.  It's
        # fine to load it from the stack too, as long as it's != r0, r1.
        lengthloc = self.rm.loc(length_box)
        self.rm.possibly_free_var(length_box)
        #
        itemsize = op.getarg(1).getint()
        maxlength = (gc_ll_descr.max_size_of_young_obj - WORD * 2) / itemsize
        self.assembler.malloc_cond_varsize(
            op.getarg(0).getint(),
            gc_ll_descr.get_nursery_free_addr(),
            gc_ll_descr.get_nursery_top_addr(),
            lengthloc, itemsize, maxlength, gcmap, arraydescr)

    prepare_op_debug_merge_point = void
    prepare_op_jit_debug = void
    prepare_op_keepalive = void
    prepare_op_enter_portal_frame = void
    prepare_op_leave_portal_frame = void

    def prepare_op_cond_call_gc_wb(self, op, fcond):
        # we force all arguments in a reg because it will be needed anyway by
        # the following gc_store. It avoids loading it twice from the memory.
        N = op.numargs()
        args = op.getarglist()
        arglocs = [self.make_sure_var_in_reg(op.getarg(i), args)
                                                              for i in range(N)]
        tmp = self.get_scratch_reg(INT, args)
        assert tmp not in arglocs
        arglocs.append(tmp)
        return arglocs

    prepare_op_cond_call_gc_wb_array = prepare_op_cond_call_gc_wb

    def prepare_op_cond_call(self, op, fcond):
        # XXX don't force the arguments to be loaded in specific
        # locations before knowing if we can take the fast path
        assert 2 <= op.numargs() <= 4 + 2
        tmpreg = self.get_scratch_reg(INT, selected_reg=r.r4)
        v = op.getarg(1)
        assert isinstance(v, Const)
        imm = self.rm.convert_to_imm(v)
        self.assembler.regalloc_mov(imm, tmpreg)
        args_so_far = []
        for i in range(2, op.numargs()):
            reg = r.argument_regs[i - 2]
            arg = op.getarg(i)
            self.make_sure_var_in_reg(arg, args_so_far, selected_reg=reg)
            args_so_far.append(arg)

        if op.type == 'v':
            # a plain COND_CALL.  Calls the function when args[0] is
            # true.  Often used just after a comparison operation.
            self.load_condition_into_cc(op.getarg(0))
            return [tmpreg]
        else:
            # COND_CALL_VALUE_I/R.  Calls the function when args[0]
            # is equal to 0 or NULL.  Returns the result from the
            # function call if done, or args[0] if it was not 0/NULL.
            # Implemented by forcing the result to live in the same
            # register as args[0], and overwriting it if we really do
            # the call.

            # Load the register for the result.  Possibly reuse 'args[0]'.
            # But the old value of args[0], if it survives, is first
            # spilled away.  We can't overwrite any of op.args[2:] here.
            args = op.getarglist()
            resloc = self.rm.force_result_in_reg(op, args[0],
                                                 forbidden_vars=args[2:])
            # Test the register for the result.
            self.assembler.mc.CMP_ri(resloc.value, 0)
            self.assembler.guard_success_cc = c.EQ
            return [tmpreg, resloc]

    prepare_op_cond_call_value_i = prepare_op_cond_call
    prepare_op_cond_call_value_r = prepare_op_cond_call

    def prepare_op_force_token(self, op, fcond):
        # XXX for now we return a regular reg
        res_loc = self.force_allocate_reg(op)
        return [res_loc]

    def prepare_op_label(self, op, fcond):
        descr = op.getdescr()
        assert isinstance(descr, TargetToken)
        inputargs = op.getarglist()
        arglocs = [None] * len(inputargs)
        #
        # we use force_spill() on the boxes that are not going to be really
        # used any more in the loop, but that are kept alive anyway
        # by being in a next LABEL's or a JUMP's argument or fail_args
        # of some guard
        position = self.rm.position
        for arg in inputargs:
            assert not isinstance(arg, Const)
            if self.longevity[arg].is_last_real_use_before(position):
                self.force_spill_var(arg)

        #
        for i in range(len(inputargs)):
            arg = inputargs[i]
            assert not isinstance(arg, Const)
            loc = self.loc(arg)
            arglocs[i] = loc
            if loc.is_core_reg() or loc.is_vfp_reg():
                self.frame_manager.mark_as_free(arg)
        #
        descr._arm_arglocs = arglocs
        descr._ll_loop_code = self.assembler.mc.currpos()
        descr._arm_clt = self.assembler.current_clt
        self.assembler.target_tokens_currently_compiling[descr] = None
        self.possibly_free_vars_for_op(op)
        #
        # if the LABEL's descr is precisely the target of the JUMP at the
        # end of the same loop, i.e. if what we are compiling is a single
        # loop that ends up jumping to this LABEL, then we can now provide
        # the hints about the expected position of the spilled variables.
        jump_op = self.final_jump_op
        if jump_op is not None and jump_op.getdescr() is descr:
            self._compute_hint_frame_locations_from_descr(descr)
        return []

    def prepare_op_guard_not_forced_2(self, op, fcond):
        self.rm.before_call(op.getfailargs(), save_all_regs=True)
        self.vfprm.before_call(op.getfailargs(), save_all_regs=True)
        fail_locs = self._prepare_guard(op)
        self.assembler.store_force_descr(op, fail_locs[1:], fail_locs[0].value)
        self.possibly_free_vars(op.getfailargs())

    def _prepare_op_call_may_force(self, op, fcond):
        return self._prepare_call(op, save_all_regs=True)

    prepare_op_call_may_force_i = _prepare_op_call_may_force
    prepare_op_call_may_force_r = _prepare_op_call_may_force
    prepare_op_call_may_force_f = _prepare_op_call_may_force
    prepare_op_call_may_force_n = _prepare_op_call_may_force

    def _prepare_op_call_release_gil(self, op, fcond):
        return self._prepare_call(op, save_all_regs=True, first_arg_index=2)

    prepare_op_call_release_gil_i = _prepare_op_call_release_gil
    prepare_op_call_release_gil_f = _prepare_op_call_release_gil
    prepare_op_call_release_gil_n = _prepare_op_call_release_gil

    def _prepare_op_call_assembler(self, op, fcond):
        locs = self.locs_for_call_assembler(op)
        tmploc = self.get_scratch_reg(INT, selected_reg=r.r0)
        resloc = self._call(op, locs + [tmploc], gc_level=2)
        return locs + [resloc, tmploc]

    prepare_op_call_assembler_i = _prepare_op_call_assembler
    prepare_op_call_assembler_r = _prepare_op_call_assembler
    prepare_op_call_assembler_f = _prepare_op_call_assembler
    prepare_op_call_assembler_n = _prepare_op_call_assembler

    prepare_op_float_add = prepare_two_regs_op
    prepare_op_float_sub = prepare_two_regs_op
    prepare_op_float_mul = prepare_two_regs_op
    prepare_op_float_truediv = prepare_two_regs_op
    prepare_op_float_lt = prepare_float_cmp
    prepare_op_float_le = prepare_float_cmp
    prepare_op_float_eq = prepare_float_cmp
    prepare_op_float_ne = prepare_float_cmp
    prepare_op_float_gt = prepare_float_cmp
    prepare_op_float_ge = prepare_float_cmp
    prepare_op_float_neg = prepare_unary_op
    prepare_op_float_abs = prepare_unary_op

    def _prepare_op_math_sqrt(self, op, fcond):
        loc = self.make_sure_var_in_reg(op.getarg(1))
        self.possibly_free_vars_for_op(op)
        self.free_temp_vars()
        res = self.vfprm.force_allocate_reg(op)
        return [loc, res]

    def prepare_op_cast_float_to_int(self, op, fcond):
        loc1 = self.make_sure_var_in_reg(op.getarg(0))
        res = self.rm.force_allocate_reg(op)
        return [loc1, res]

    def prepare_op_cast_int_to_float(self, op, fcond):
        loc1 = self.make_sure_var_in_reg(op.getarg(0))
        res = self.vfprm.force_allocate_reg(op)
        return [loc1, res]

    def prepare_force_spill(self, op, fcond):
        self.force_spill_var(op.getarg(0))
        return []

    prepare_op_convert_float_bytes_to_longlong = prepare_unary_op
    prepare_op_convert_longlong_bytes_to_float = prepare_unary_op

    #def prepare_op_read_timestamp(self, op, fcond):
    #    loc = self.get_scratch_reg(INT)
    #    res = self.vfprm.force_allocate_reg(op)
    #    return [loc, res]

    def prepare_op_cast_float_to_singlefloat(self, op, fcond):
        loc1 = self.make_sure_var_in_reg(op.getarg(0))
        res = self.force_allocate_reg(op)
        return [loc1, res]

    def prepare_op_cast_singlefloat_to_float(self, op, fcond):
        loc1 = self.make_sure_var_in_reg(op.getarg(0))
        res = self.force_allocate_reg(op)
        return [loc1, res]


def notimplemented(self, op, fcond):
    print("[ARM/regalloc] %s not implemented" % op.getopname())
    raise NotImplementedError(op)


operations = [notimplemented] * (rop._LAST + 1)


for key, value in rop.__dict__.items():
    key = key.lower()
    if key.startswith('_'):
        continue
    methname = 'prepare_op_%s' % key
    if hasattr(Regalloc, methname):
        func = getattr(Regalloc, methname).im_func
        operations[value] = func



import pytest
import sys
import os
from commands import getoutput
from pypy.tool.release.smartstrip import smartstrip

@pytest.fixture
def exe(tmpdir):
    src = tmpdir.join("myprog.c")
    src.write("""
    int foo(int a, int b) {
        return a+b;
    }
    int main(void) { }
    """)
    exe = tmpdir.join("myprog")
    ret = os.system("gcc -o %s %s" % (exe, src))
    assert ret == 0
    return exe

def info_symbol(exe, symbol):
    out = getoutput("gdb %s -ex 'info symbol %s' -ex 'quit'" % (exe, symbol))
    lines = out.splitlines()
    return lines[-1]

@pytest.mark.skipif(sys.platform == 'win32',
                    reason='strip not supported on windows')
class TestSmarStrip(object):

    def test_info_symbol(self, exe):
        info = info_symbol(exe, "foo")
        assert info == "foo in section .text"

    def test_strip(self, exe):
        smartstrip(exe, keep_debug=False)
        info = info_symbol(exe, "foo")
        assert info.startswith("No symbol table is loaded")

    @pytest.mark.skipif(sys.platform != 'linux2',
                        reason='keep_debug not supported')
    def test_keep_debug(self, exe, tmpdir):
        smartstrip(exe, keep_debug=True)
        debug = tmpdir.join("myprog.debug")
        assert debug.check(file=True)
        perm = debug.stat().mode & 0777
        assert perm & 0111 == 0 # 'x' bit not set
        #
        info = info_symbol(exe, "foo")
        assert info == "foo in section .text of %s" % exe
        #
        debug.remove()
        info = info_symbol(exe, "foo")
        assert info.startswith("No symbol table is loaded")



from rpython.jit.backend.arm.test.support import skip_unless_run_slow_tests
skip_unless_run_slow_tests()

from rpython.jit.backend.llsupport.test.zrpy_releasegil_test import ReleaseGILTests


class TestShadowStack(ReleaseGILTests):
    gcrootfinder = "shadowstack"





import sys
from rpython.tool.pairtype import pairtype
from rpython.flowspace.model import Constant
from rpython.rtyper.rdict import AbstractDictRepr, AbstractDictIteratorRepr
from rpython.rtyper.lltypesystem import lltype, llmemory, rffi
from rpython.rlib import objectmodel, jit, rgc, types
from rpython.rlib.signature import signature
from rpython.rlib.objectmodel import specialize, likely, not_rpython
from rpython.rtyper.debug import ll_assert
from rpython.rlib.rarithmetic import r_uint, intmask
from rpython.rtyper import rmodel
from rpython.rtyper.error import TyperError
from rpython.rtyper.annlowlevel import llhelper


# ____________________________________________________________
#
#  generic implementation of RPython dictionary, with parametric DICTKEY and
#  DICTVALUE types. The basic implementation is a sparse array of indexes
#  plus a dense array of structs that contain keys and values. struct looks
#  like that:
#
#
#    struct dictentry {
#        DICTKEY key;
#        DICTVALUE value;
#        long f_hash;        # (optional) key hash, if hard to recompute
#        bool f_valid;      # (optional) the entry is filled
#    }
#
#    struct dicttable {
#        int num_live_items;
#        int num_ever_used_items;
#        int resize_counter;
#        {byte, short, int, long} *indexes;
#        dictentry *entries;
#        lookup_function_no; # one of the four possible functions for different
#                       # size dicts; the rest of the word is a counter for how
#                       # many 'entries' at the start are known to be deleted
#        (Function DICTKEY, DICTKEY -> bool) *fnkeyeq;
#        (Function DICTKEY -> int) *fnkeyhash;
#    }
#
#

@jit.look_inside_iff(lambda d, key, hash, flag: jit.isvirtual(d))
@jit.oopspec('ordereddict.lookup(d, key, hash, flag)')
def ll_call_lookup_function(d, key, hash, flag):
    while True:
        fun = d.lookup_function_no & FUNC_MASK
        # This likely() here forces gcc to compile the check for fun==FUNC_BYTE
        # first.  Otherwise, this is a regular switch and gcc (at least 4.7)
        # compiles this as a series of checks, with the FUNC_BYTE case last.
        # It sounds minor, but it is worth 6-7% on a PyPy microbenchmark.
        if likely(fun == FUNC_BYTE):
            return ll_dict_lookup(d, key, hash, flag, TYPE_BYTE)
        elif fun == FUNC_SHORT:
            return ll_dict_lookup(d, key, hash, flag, TYPE_SHORT)
        elif IS_64BIT and fun == FUNC_INT:
            return ll_dict_lookup(d, key, hash, flag, TYPE_INT)
        elif fun == FUNC_LONG:
            return ll_dict_lookup(d, key, hash, flag, TYPE_LONG)
        else:
            ll_dict_create_initial_index(d)
            # then, retry

def get_ll_dict(DICTKEY, DICTVALUE, get_custom_eq_hash=None, DICT=None,
                ll_fasthash_function=None, ll_hash_function=None,
                ll_eq_function=None, method_cache={}, simple_hash_eq=False,
                dummykeyobj=None, dummyvalueobj=None, rtyper=None):
    # get the actual DICT type. if DICT is None, it's created, otherwise
    # forward reference is becoming DICT
    if DICT is None:
        DICT = lltype.GcForwardReference()
    # compute the shape of the DICTENTRY structure
    entryfields = []
    entrymeths = {
        'allocate': lltype.typeMethod(_ll_malloc_entries),
        'delete': _ll_free_entries,
        'must_clear_key':   (isinstance(DICTKEY, lltype.Ptr)
                             and DICTKEY._needsgc()),
        'must_clear_value': (isinstance(DICTVALUE, lltype.Ptr)
                             and DICTVALUE._needsgc()),
        }
    if getattr(ll_eq_function, 'no_direct_compare', False):
        entrymeths['no_direct_compare'] = True

    # * the key
    entryfields.append(("key", DICTKEY))

    # * the state of the entry - trying to encode it as dummy objects
    if dummykeyobj:
        # all the state can be encoded in the key
        entrymeths['dummy_obj'] = dummykeyobj
        entrymeths['valid'] = ll_valid_from_key
        entrymeths['mark_deleted'] = ll_mark_deleted_in_key
        # the key is overwritten by 'dummy' when the entry is deleted
        entrymeths['must_clear_key'] = False

    elif dummyvalueobj:
        # all the state can be encoded in the value
        entrymeths['dummy_obj'] = dummyvalueobj
        entrymeths['valid'] = ll_valid_from_value
        entrymeths['mark_deleted'] = ll_mark_deleted_in_value
        # value is overwritten by 'dummy' when entry is deleted
        entrymeths['must_clear_value'] = False

    else:
        # we need a flag to know if the entry was ever used
        entryfields.append(("f_valid", lltype.Bool))
        entrymeths['valid'] = ll_valid_from_flag
        entrymeths['mark_deleted'] = ll_mark_deleted_in_flag

    # * the value
    entryfields.append(("value", DICTVALUE))

    if simple_hash_eq:
        assert get_custom_eq_hash is not None
        entrymeths['entry_hash'] = ll_hash_custom_fast
    elif ll_fasthash_function is None:
        entryfields.append(("f_hash", lltype.Signed))
        entrymeths['entry_hash'] = ll_hash_from_cache
    else:
        entrymeths['entry_hash'] = ll_hash_recomputed
        entrymeths['fasthashfn'] = ll_fasthash_function

    # Build the lltype data structures
    DICTENTRY = lltype.Struct("odictentry", *entryfields)
    DICTENTRYARRAY = lltype.GcArray(DICTENTRY,
                                    adtmeths=entrymeths)
    fields =          [ ("num_live_items", lltype.Signed),
                        ("num_ever_used_items", lltype.Signed),
                        ("resize_counter", lltype.Signed),
                        ("indexes", llmemory.GCREF),
                        ("lookup_function_no", lltype.Signed),
                        ("entries", lltype.Ptr(DICTENTRYARRAY)) ]
    if get_custom_eq_hash is not None:
        r_rdict_eqfn, r_rdict_hashfn = get_custom_eq_hash()
        fields.extend([ ("fnkeyeq", r_rdict_eqfn.lowleveltype),
                        ("fnkeyhash", r_rdict_hashfn.lowleveltype) ])
        adtmeths = {
            'keyhash':        ll_keyhash_custom,
            'keyeq':          ll_keyeq_custom,
            'r_rdict_eqfn':   r_rdict_eqfn,
            'r_rdict_hashfn': r_rdict_hashfn,
            'paranoia':       not simple_hash_eq,
            }
    else:
        # figure out which functions must be used to hash and compare
        ll_keyhash = ll_hash_function
        ll_keyeq = ll_eq_function
        ll_keyhash = lltype.staticAdtMethod(ll_keyhash)
        if ll_keyeq is not None:
            ll_keyeq = lltype.staticAdtMethod(ll_keyeq)
        adtmeths = {
            'keyhash':  ll_keyhash,
            'keyeq':    ll_keyeq,
            'paranoia': False,
            }
    adtmeths['KEY']   = DICTKEY
    adtmeths['VALUE'] = DICTVALUE
    adtmeths['lookup_function'] = lltype.staticAdtMethod(ll_call_lookup_function)
    adtmeths['allocate'] = lltype.typeMethod(_ll_malloc_dict)

    DICT.become(lltype.GcStruct("dicttable", adtmeths=adtmeths,
                                *fields))
    return DICT


class OrderedDictRepr(AbstractDictRepr):

    def __init__(self, rtyper, key_repr, value_repr, dictkey, dictvalue,
                 custom_eq_hash=None, force_non_null=False, simple_hash_eq=False):
        #assert not force_non_null
        self.rtyper = rtyper
        self.finalized = False
        self.DICT = lltype.GcForwardReference()
        self.lowleveltype = lltype.Ptr(self.DICT)
        self.custom_eq_hash = custom_eq_hash is not None
        self.simple_hash_eq = simple_hash_eq
        if not isinstance(key_repr, rmodel.Repr):  # not computed yet, done by setup()
            assert callable(key_repr)
            self._key_repr_computer = key_repr
        else:
            self.external_key_repr, self.key_repr = self.pickkeyrepr(key_repr)
        if not isinstance(value_repr, rmodel.Repr):  # not computed yet, done by setup()
            assert callable(value_repr)
            self._value_repr_computer = value_repr
        else:
            self.external_value_repr, self.value_repr = self.pickrepr(value_repr)
        self.dictkey = dictkey
        self.dictvalue = dictvalue
        self.dict_cache = {}
        self._custom_eq_hash_repr = custom_eq_hash
        # setup() needs to be called to finish this initialization

    def _externalvsinternal(self, rtyper, item_repr):
        return rmodel.externalvsinternal(self.rtyper, item_repr)

    def _setup_repr(self):
        if 'key_repr' not in self.__dict__:
            key_repr = self._key_repr_computer()
            self.external_key_repr, self.key_repr = self.pickkeyrepr(key_repr)
        if 'value_repr' not in self.__dict__:
            self.external_value_repr, self.value_repr = rmodel.externalvsinternal(self.rtyper, self._value_repr_computer(), gcref=True)
        if isinstance(self.DICT, lltype.GcForwardReference):
            DICTKEY = self.key_repr.lowleveltype
            DICTVALUE = self.value_repr.lowleveltype
            # * we need an explicit flag if the key and the value is not
            #   able to store dummy values
            s_key   = self.dictkey.s_value
            s_value = self.dictvalue.s_value
            kwd = {}
            if self.custom_eq_hash:
                self.r_rdict_eqfn, self.r_rdict_hashfn = (
                    self._custom_eq_hash_repr())
                kwd['get_custom_eq_hash'] = self._custom_eq_hash_repr
                kwd['simple_hash_eq'] = self.simple_hash_eq
            else:
                kwd['ll_hash_function'] = self.key_repr.get_ll_hash_function()
                kwd['ll_eq_function'] = self.key_repr.get_ll_eq_function()
                kwd['ll_fasthash_function'] = self.key_repr.get_ll_fasthash_function()
            kwd['dummykeyobj'] = self.key_repr.get_ll_dummyval_obj(self.rtyper,
                                                                   s_key)
            kwd['dummyvalueobj'] = self.value_repr.get_ll_dummyval_obj(
                self.rtyper, s_value)
            get_ll_dict(DICTKEY, DICTVALUE, DICT=self.DICT,
                        rtyper=self.rtyper, **kwd)


    def convert_const(self, dictobj):
        from rpython.rtyper.lltypesystem import llmemory
        # get object from bound dict methods
        #dictobj = getattr(dictobj, '__self__', dictobj)
        if dictobj is None:
            return lltype.nullptr(self.DICT)
        if not isinstance(dictobj, (dict, objectmodel.r_dict)):
            raise TypeError("expected a dict: %r" % (dictobj,))
        try:
            key = Constant(dictobj)
            return self.dict_cache[key]
        except KeyError:
            self.setup()
            self.setup_final()
            l_dict = ll_newdict_size(self.DICT, len(dictobj))
            ll_no_initial_index(l_dict)
            self.dict_cache[key] = l_dict
            r_key = self.key_repr
            if r_key.lowleveltype == llmemory.Address:
                raise TypeError("No prebuilt dicts of address keys")
            r_value = self.value_repr
            if isinstance(dictobj, objectmodel.r_dict):

                if self.r_rdict_eqfn.lowleveltype != lltype.Void:
                    l_fn = self.r_rdict_eqfn.convert_const(dictobj.key_eq)
                    l_dict.fnkeyeq = l_fn
                if self.r_rdict_hashfn.lowleveltype != lltype.Void:
                    l_fn = self.r_rdict_hashfn.convert_const(dictobj.key_hash)
                    l_dict.fnkeyhash = l_fn

                for dictkeycontainer, dictvalue in dictobj._dict.items():
                    llkey = r_key.convert_const(dictkeycontainer.key)
                    llvalue = r_value.convert_const(dictvalue)
                    _ll_dict_insert_no_index(l_dict, llkey, llvalue)
                return l_dict

            else:
                for dictkey, dictvalue in dictobj.items():
                    llkey = r_key.convert_const(dictkey)
                    llvalue = r_value.convert_const(dictvalue)
                    _ll_dict_insert_no_index(l_dict, llkey, llvalue)
                return l_dict

    def rtype_len(self, hop):
        v_dict, = hop.inputargs(self)
        return hop.gendirectcall(ll_dict_len, v_dict)

    def rtype_bool(self, hop):
        v_dict, = hop.inputargs(self)
        return hop.gendirectcall(ll_dict_bool, v_dict)

    def make_iterator_repr(self, *variant):
        return DictIteratorRepr(self, *variant)

    def rtype_method_get(self, hop):
        if hop.nb_args == 3:
            v_dict, v_key, v_default = hop.inputargs(self, self.key_repr,
                                                     self.value_repr)
        else:
            v_dict, v_key = hop.inputargs(self, self.key_repr)
            v_default = hop.inputconst(self.value_repr, None)
        hop.exception_cannot_occur()
        v_res = hop.gendirectcall(ll_dict_get, v_dict, v_key, v_default)
        return self.recast_value(hop.llops, v_res)

    def rtype_method_setdefault(self, hop):
        v_dict, v_key, v_default = hop.inputargs(self, self.key_repr,
                                                 self.value_repr)
        hop.exception_cannot_occur()
        v_res = hop.gendirectcall(ll_dict_setdefault, v_dict, v_key, v_default)
        return self.recast_value(hop.llops, v_res)

    def rtype_method_copy(self, hop):
        v_dict, = hop.inputargs(self)
        hop.exception_cannot_occur()
        return hop.gendirectcall(ll_dict_copy, v_dict)

    def rtype_method_update(self, hop):
        v_dic1, v_dic2 = hop.inputargs(self, self)
        hop.exception_cannot_occur()
        return hop.gendirectcall(ll_dict_update, v_dic1, v_dic2)

    def rtype_method__prepare_dict_update(self, hop):
        v_dict, v_num = hop.inputargs(self, lltype.Signed)
        hop.exception_cannot_occur()
        hop.gendirectcall(ll_prepare_dict_update, v_dict, v_num)

    def _rtype_method_kvi(self, hop, ll_func, extraarg=None):
        v_dic, = hop.inputargs(self)
        r_list = hop.r_result
        cLIST = hop.inputconst(lltype.Void, r_list.lowleveltype.TO)
        hop.exception_cannot_occur()
        if extraarg:
            c_extraarg = hop.inputconst(lltype.Void, extraarg)
        else:
            c_extraarg = hop.inputconst(lltype.Void, None)
        return hop.gendirectcall(ll_func, cLIST, v_dic, c_extraarg)

    def rtype_method_keys(self, hop):
        return self._rtype_method_kvi(hop, ll_dict_keys)

    def rtype_method_values(self, hop):
        return self._rtype_method_kvi(hop, ll_dict_values)

    def rtype_method_items(self, hop):
        EXTERNAL_ELEM = hop.r_result.external_item_repr.lowleveltype
        return self._rtype_method_kvi(hop, ll_dict_items, EXTERNAL_ELEM)

    def rtype_bltn_list(self, hop):
        return self._rtype_method_kvi(hop, ll_dict_keys)

    def rtype_method_iterkeys(self, hop):
        hop.exception_cannot_occur()
        return DictIteratorRepr(self, "keys").newiter(hop)

    def rtype_method_itervalues(self, hop):
        hop.exception_cannot_occur()
        return DictIteratorRepr(self, "values").newiter(hop)

    def rtype_method_iteritems(self, hop):
        hop.exception_cannot_occur()
        return DictIteratorRepr(self, "items").newiter(hop)

    def rtype_method_iterkeys_with_hash(self, hop):
        v_dic, = hop.inputargs(self)
        hop.exception_is_here()
        hop.gendirectcall(ll_ensure_indexes, v_dic)
        return DictIteratorRepr(self, "keys_with_hash").newiter(hop)

    def rtype_method_iteritems_with_hash(self, hop):
        v_dic, = hop.inputargs(self)
        hop.exception_is_here()
        hop.gendirectcall(ll_ensure_indexes, v_dic)
        return DictIteratorRepr(self, "items_with_hash").newiter(hop)

    def rtype_method_clear(self, hop):
        v_dict, = hop.inputargs(self)
        hop.exception_cannot_occur()
        return hop.gendirectcall(ll_dict_clear, v_dict)

    def rtype_method_popitem(self, hop):
        v_dict, = hop.inputargs(self)
        r_tuple = hop.r_result
        cTUPLE = hop.inputconst(lltype.Void, r_tuple.lowleveltype)
        hop.exception_is_here()
        return hop.gendirectcall(ll_dict_popitem, cTUPLE, v_dict)

    def rtype_method_pop(self, hop):
        if hop.nb_args == 2:
            v_args = hop.inputargs(self, self.key_repr)
            target = ll_dict_pop
        elif hop.nb_args == 3:
            v_args = hop.inputargs(self, self.key_repr, self.value_repr)
            target = ll_dict_pop_default
        hop.exception_is_here()
        v_res = hop.gendirectcall(target, *v_args)
        return self.recast_value(hop.llops, v_res)

    def rtype_method_contains_with_hash(self, hop):
        v_dict, v_key, v_hash = hop.inputargs(self, self.key_repr,
                                              lltype.Signed)
        hop.exception_is_here()
        return hop.gendirectcall(ll_dict_contains_with_hash,
                                 v_dict, v_key, v_hash)

    def rtype_method_setitem_with_hash(self, hop):
        v_dict, v_key, v_hash, v_value = hop.inputargs(
            self, self.key_repr, lltype.Signed, self.value_repr)
        if self.custom_eq_hash:
            hop.exception_is_here()
        else:
            hop.exception_cannot_occur()
        hop.gendirectcall(ll_dict_setitem_with_hash,
                          v_dict, v_key, v_hash, v_value)

    def rtype_method_getitem_with_hash(self, hop):
        v_dict, v_key, v_hash = hop.inputargs(
            self, self.key_repr, lltype.Signed)
        if not self.custom_eq_hash:
            hop.has_implicit_exception(KeyError)  # record that we know about it
        hop.exception_is_here()
        v_res = hop.gendirectcall(ll_dict_getitem_with_hash,
                                  v_dict, v_key, v_hash)
        return self.recast_value(hop.llops, v_res)

    def rtype_method_delitem_with_hash(self, hop):
        v_dict, v_key, v_hash = hop.inputargs(
            self, self.key_repr, lltype.Signed)
        if not self.custom_eq_hash:
            hop.has_implicit_exception(KeyError)  # record that we know about it
        hop.exception_is_here()
        hop.gendirectcall(ll_dict_delitem_with_hash, v_dict, v_key, v_hash)

    def rtype_method_delitem_if_value_is(self, hop):
        v_dict, v_key, v_value = hop.inputargs(
            self, self.key_repr, self.value_repr)
        hop.exception_cannot_occur()
        hop.gendirectcall(ll_dict_delitem_if_value_is, v_dict, v_key, v_value)

    def rtype_method_move_to_end(self, hop):
        v_dict, v_key, v_last = hop.inputargs(
            self, self.key_repr, lltype.Bool)
        if not self.custom_eq_hash:
            hop.has_implicit_exception(KeyError)  # record that we know about it
        hop.exception_is_here()
        hop.gendirectcall(ll_dict_move_to_end, v_dict, v_key, v_last)


class __extend__(pairtype(OrderedDictRepr, rmodel.Repr)):

    def rtype_getitem((r_dict, r_key), hop):
        v_dict, v_key = hop.inputargs(r_dict, r_dict.key_repr)
        if not r_dict.custom_eq_hash:
            hop.has_implicit_exception(KeyError)   # record that we know about it
        hop.exception_is_here()
        v_res = hop.gendirectcall(ll_dict_getitem, v_dict, v_key)
        return r_dict.recast_value(hop.llops, v_res)

    def rtype_delitem((r_dict, r_key), hop):
        v_dict, v_key = hop.inputargs(r_dict, r_dict.key_repr)
        if not r_dict.custom_eq_hash:
            hop.has_implicit_exception(KeyError)   # record that we know about it
        hop.exception_is_here()
        hop.gendirectcall(ll_dict_delitem, v_dict, v_key)

    def rtype_setitem((r_dict, r_key), hop):
        v_dict, v_key, v_value = hop.inputargs(r_dict, r_dict.key_repr, r_dict.value_repr)
        if r_dict.custom_eq_hash:
            hop.exception_is_here()
        else:
            hop.exception_cannot_occur()
        hop.gendirectcall(ll_dict_setitem, v_dict, v_key, v_value)

    def rtype_contains((r_dict, r_key), hop):
        v_dict, v_key = hop.inputargs(r_dict, r_dict.key_repr)
        hop.exception_is_here()
        return hop.gendirectcall(ll_dict_contains, v_dict, v_key)

class __extend__(pairtype(OrderedDictRepr, OrderedDictRepr)):
    def convert_from_to((r_dict1, r_dict2), v, llops):
        # check that we don't convert from Dicts with
        # different key/value types
        if r_dict1.dictkey is None or r_dict2.dictkey is None:
            return NotImplemented
        if r_dict1.dictkey is not r_dict2.dictkey:
            return NotImplemented
        if r_dict1.dictvalue is None or r_dict2.dictvalue is None:
            return NotImplemented
        if r_dict1.dictvalue is not r_dict2.dictvalue:
            return NotImplemented
        return v

# ____________________________________________________________
#
#  Low-level methods.  These can be run for testing, but are meant to
#  be direct_call'ed from rtyped flow graphs, which means that they will
#  get flowed and annotated, mostly with SomePtr.

DICTINDEX_LONG = lltype.Ptr(lltype.GcArray(lltype.Unsigned))
DICTINDEX_INT = lltype.Ptr(lltype.GcArray(rffi.UINT))
DICTINDEX_SHORT = lltype.Ptr(lltype.GcArray(rffi.USHORT))
DICTINDEX_BYTE = lltype.Ptr(lltype.GcArray(rffi.UCHAR))

IS_64BIT = sys.maxint != 2 ** 31 - 1

if IS_64BIT:
    FUNC_SHIFT = 3
    FUNC_MASK  = 0x07  # three bits
    FUNC_BYTE, FUNC_SHORT, FUNC_INT, FUNC_LONG, FUNC_MUST_REINDEX = range(5)
else:
    FUNC_SHIFT = 2
    FUNC_MASK  = 0x03  # two bits
    FUNC_BYTE, FUNC_SHORT, FUNC_LONG, FUNC_MUST_REINDEX = range(4)
TYPE_BYTE  = rffi.UCHAR
TYPE_SHORT = rffi.USHORT
TYPE_INT   = rffi.UINT
TYPE_LONG  = lltype.Unsigned

def ll_no_initial_index(d):
    # Used when making new empty dicts, and when translating prebuilt dicts.
    # Remove the index completely.  A dictionary must always have an
    # index unless it is freshly created or freshly translated.  Most
    # dict operations start with ll_call_lookup_function(), which will
    # recompute the hashes and create the index.
    ll_assert(d.num_live_items == d.num_ever_used_items, 
         "ll_no_initial_index(): dict already in use")
    d.lookup_function_no = FUNC_MUST_REINDEX
    d.indexes = lltype.nullptr(llmemory.GCREF.TO)

def ll_malloc_indexes_and_choose_lookup(d, n):
    # keep in sync with ll_clear_indexes() below
    if n <= 256:
        d.indexes = lltype.cast_opaque_ptr(llmemory.GCREF,
                                           lltype.malloc(DICTINDEX_BYTE.TO, n,
                                                         zero=True))
        d.lookup_function_no = FUNC_BYTE
    elif n <= 65536:
        d.indexes = lltype.cast_opaque_ptr(llmemory.GCREF,
                                           lltype.malloc(DICTINDEX_SHORT.TO, n,
                                                         zero=True))
        d.lookup_function_no = FUNC_SHORT
    elif IS_64BIT and n <= 2 ** 32:
        d.indexes = lltype.cast_opaque_ptr(llmemory.GCREF,
                                           lltype.malloc(DICTINDEX_INT.TO, n,
                                                         zero=True))
        d.lookup_function_no = FUNC_INT
    else:
        d.indexes = lltype.cast_opaque_ptr(llmemory.GCREF,
                                           lltype.malloc(DICTINDEX_LONG.TO, n,
                                                         zero=True))
        d.lookup_function_no = FUNC_LONG

def ll_clear_indexes(d, n):
    fun = d.lookup_function_no & FUNC_MASK
    d.lookup_function_no = fun
    if fun == FUNC_BYTE:
        rgc.ll_arrayclear(lltype.cast_opaque_ptr(DICTINDEX_BYTE, d.indexes))
    elif fun == FUNC_SHORT:
        rgc.ll_arrayclear(lltype.cast_opaque_ptr(DICTINDEX_SHORT, d.indexes))
    elif IS_64BIT and fun == FUNC_INT:
        rgc.ll_arrayclear(lltype.cast_opaque_ptr(DICTINDEX_INT, d.indexes))
    elif fun == FUNC_LONG:
        rgc.ll_arrayclear(lltype.cast_opaque_ptr(DICTINDEX_LONG, d.indexes))
    else:
        assert False

@objectmodel.always_inline
def _ll_write_indexes(d, i, value, T):
    INDEXES = _ll_ptr_to_array_of(T)
    indexes = lltype.cast_opaque_ptr(INDEXES, d.indexes)
    cast_value = rffi.cast(T, value)
    ll_assert(intmask(cast_value) == value, "indexes are a too small type of array")
    indexes[i] = cast_value

@jit.dont_look_inside
def ll_call_insert_clean_function(d, hash, i):
    assert i >= 0
    fun = d.lookup_function_no & FUNC_MASK
    if fun == FUNC_BYTE:
        ll_dict_store_clean(d, hash, i, TYPE_BYTE)
    elif fun == FUNC_SHORT:
        ll_dict_store_clean(d, hash, i, TYPE_SHORT)
    elif IS_64BIT and fun == FUNC_INT:
        ll_dict_store_clean(d, hash, i, TYPE_INT)
    elif fun == FUNC_LONG:
        ll_dict_store_clean(d, hash, i, TYPE_LONG)
    else:
        # can't be still FUNC_MUST_REINDEX here
        ll_assert(False, "ll_call_insert_clean_function(): invalid lookup_fun")
        assert False

def ll_call_delete_by_entry_index(d, hash, i, replace_with):
    # only called from _ll_dict_del, whose @jit.look_inside_iff
    # condition should control when we get inside here with the jit
    fun = d.lookup_function_no & FUNC_MASK
    if fun == FUNC_BYTE:
        ll_dict_delete_by_entry_index(d, hash, i, replace_with, TYPE_BYTE)
    elif fun == FUNC_SHORT:
        ll_dict_delete_by_entry_index(d, hash, i, replace_with, TYPE_SHORT)
    elif IS_64BIT and fun == FUNC_INT:
        ll_dict_delete_by_entry_index(d, hash, i, replace_with, TYPE_INT)
    elif fun == FUNC_LONG:
        ll_dict_delete_by_entry_index(d, hash, i, replace_with, TYPE_LONG)
    else:
        # can't be still FUNC_MUST_REINDEX here
        ll_assert(False, "ll_call_delete_by_entry_index(): invalid lookup_fun")
        assert False

def ll_valid_from_flag(entries, i):
    return entries[i].f_valid

def ll_valid_from_key(entries, i):
    ENTRIES = lltype.typeOf(entries).TO
    dummy = ENTRIES.dummy_obj.ll_dummy_value
    return entries[i].key != dummy

def ll_valid_from_value(entries, i):
    ENTRIES = lltype.typeOf(entries).TO
    dummy = ENTRIES.dummy_obj.ll_dummy_value
    return entries[i].value != dummy

def ll_mark_deleted_in_flag(entries, i):
    entries[i].f_valid = False

def ll_mark_deleted_in_key(entries, i):
    ENTRIES = lltype.typeOf(entries).TO
    dummy = ENTRIES.dummy_obj.ll_dummy_value
    entries[i].key = dummy

def ll_mark_deleted_in_value(entries, i):
    ENTRIES = lltype.typeOf(entries).TO
    dummy = ENTRIES.dummy_obj.ll_dummy_value
    entries[i].value = dummy

@signature(types.any(), types.any(), types.int(), returns=types.any())
def ll_hash_from_cache(entries, d, i):
    return entries[i].f_hash

@signature(types.any(), types.any(), types.int(), returns=types.any())
def ll_hash_recomputed(entries, d, i):
    ENTRIES = lltype.typeOf(entries).TO
    return ENTRIES.fasthashfn(entries[i].key)

@signature(types.any(), types.any(), types.int(), returns=types.any())
def ll_hash_custom_fast(entries, d, i):
    DICT = lltype.typeOf(d).TO
    key = entries[i].key
    return objectmodel.hlinvoke(DICT.r_rdict_hashfn, d.fnkeyhash, key)

def ll_keyhash_custom(d, key):
    DICT = lltype.typeOf(d).TO
    return objectmodel.hlinvoke(DICT.r_rdict_hashfn, d.fnkeyhash, key)

def ll_keyeq_custom(d, key1, key2):
    DICT = lltype.typeOf(d).TO
    return objectmodel.hlinvoke(DICT.r_rdict_eqfn, d.fnkeyeq, key1, key2)

def ll_dict_len(d):
    return d.num_live_items

def ll_dict_bool(d):
    # check if a dict is True, allowing for None
    return bool(d) and d.num_live_items != 0

def ll_dict_getitem(d, key):
    return ll_dict_getitem_with_hash(d, key, d.keyhash(key))

def ll_dict_getitem_with_hash(d, key, hash):
    index = d.lookup_function(d, key, hash, FLAG_LOOKUP)
    if index >= 0:
        return d.entries[index].value
    else:
        raise KeyError

def ll_dict_setitem(d, key, value):
    ll_dict_setitem_with_hash(d, key, d.keyhash(key), value)

def ll_dict_setitem_with_hash(d, key, hash, value):
    index = d.lookup_function(d, key, hash, FLAG_STORE)
    _ll_dict_setitem_lookup_done(d, key, value, hash, index)

# It may be safe to look inside always, it has a few branches though, and their
# frequencies needs to be investigated.
@jit.look_inside_iff(lambda d, key, value, hash, i: jit.isvirtual(d) and jit.isconstant(key))
def _ll_dict_setitem_lookup_done(d, key, value, hash, i):
    ENTRY = lltype.typeOf(d.entries).TO.OF
    if i >= 0:
        entry = d.entries[i]
        entry.value = value
    else:
        reindexed = False
        if len(d.entries) == d.num_ever_used_items:
            try:
                reindexed = ll_dict_grow(d)
            except:
                _ll_dict_rescue(d)
                raise
        rc = d.resize_counter - 3
        if rc <= 0:
            try:
                ll_dict_resize(d)
                reindexed = True
            except:
                _ll_dict_rescue(d)
                raise
            rc = d.resize_counter - 3
            ll_assert(rc > 0, "ll_dict_resize failed?")
        if reindexed:
            ll_call_insert_clean_function(d, hash, d.num_ever_used_items)
        #
        d.resize_counter = rc
        entry = d.entries[d.num_ever_used_items]
        entry.key = key
        entry.value = value
        if hasattr(ENTRY, 'f_hash'):
            entry.f_hash = hash
        if hasattr(ENTRY, 'f_valid'):
            entry.f_valid = True
        d.num_ever_used_items += 1
        d.num_live_items += 1

@jit.dont_look_inside
def _ll_dict_rescue(d):
    # MemoryError situation!  The 'indexes' contains an invalid entry
    # at this point.  But we can call ll_dict_reindex() with the
    # following arguments, ensuring no further malloc occurs.
    ll_dict_reindex(d, _ll_len_of_d_indexes(d))
_ll_dict_rescue._dont_inline_ = True

@not_rpython
def _ll_dict_insert_no_index(d, key, value):
    # never translated
    ENTRY = lltype.typeOf(d.entries).TO.OF
    entry = d.entries[d.num_ever_used_items]
    entry.key = key
    entry.value = value
    # note that f_hash is left uninitialized in prebuilt dicts
    if hasattr(ENTRY, 'f_valid'):
        entry.f_valid = True
    d.num_ever_used_items += 1
    d.num_live_items += 1
    rc = d.resize_counter - 3
    d.resize_counter = rc

@jit.dont_look_inside
def ll_len_of_d_indexes(d):
    return _ll_len_of_d_indexes(d)

def _ll_len_of_d_indexes(d):
    # xxx Haaaack: returns len(d.indexes).  Works independently of
    # the exact type pointed to by d, using a forced cast...
    # Must only be called by @jit.dont_look_inside functions.
    return lltype.length_of_simple_gcarray_from_opaque(d.indexes)

def _overallocate_entries_len(baselen):
    # This over-allocates proportional to the list size, making room
    # for additional growth.  This over-allocates slightly more eagerly
    # than with regular lists.  The idea is that there are many more
    # lists than dicts around in PyPy, and dicts of 5 to 8 items are
    # not that rare (so a single jump from 0 to 8 is a good idea).
    # The growth pattern is:  0, 8, 17, 27, 38, 50, 64, 80, 98, ...
    newsize = baselen + (baselen >> 3)
    return newsize + 8

@jit.look_inside_iff(lambda d: jit.isvirtual(d))
def ll_dict_grow(d):
    # note: this @jit.look_inside_iff is here to inline the three lines
    # at the end of this function.  It's important because dicts start
    # with a length-zero 'd.entries' which must be grown as soon as we
    # insert an element.
    if d.num_live_items < d.num_ever_used_items // 2:
        # At least 50% of the allocated entries are dead, so perform a
        # compaction. If ll_dict_remove_deleted_items detects that over
        # 75% of allocated entries are dead, then it will also shrink the
        # memory allocated at the same time as doing a compaction.
        ll_dict_remove_deleted_items(d)
        return True

    new_allocated = _overallocate_entries_len(len(d.entries))

    # Detect a relatively rare case where the indexes numeric type is too
    # small to store all the entry indexes: there would be 'new_allocated'
    # entries, which may in corner cases be larger than 253 even though we
    # have single bytes in 'd.indexes' (and the same for the larger
    # boundaries).  The 'd.indexes' hashtable is never more than 2/3rd
    # full, so we know that 'd.num_live_items' should be at most 2/3 * 256
    # (or 65536 or etc.) so after the ll_dict_remove_deleted_items() below
    # at least 1/3rd items in 'd.entries' are free.
    if _ll_dict_entries_size_too_big(d, new_allocated):
        ll_dict_remove_deleted_items(d)
        assert d.num_live_items == d.num_ever_used_items
        return True

    newitems = lltype.malloc(lltype.typeOf(d).TO.entries.TO, new_allocated)
    rgc.ll_arraycopy(d.entries, newitems, 0, 0, len(d.entries))
    d.entries = newitems
    return False

def _ll_dict_entries_size_too_big(d, new_allocated):
    fun = d.lookup_function_no & FUNC_MASK
    if fun == FUNC_BYTE:
        assert d.num_live_items < ((1 << 8) - MIN_INDEXES_MINUS_ENTRIES)
        return new_allocated > ((1 << 8) - MIN_INDEXES_MINUS_ENTRIES)
    elif fun == FUNC_SHORT:
        assert d.num_live_items < ((1 << 16) - MIN_INDEXES_MINUS_ENTRIES)
        return new_allocated > ((1 << 16) - MIN_INDEXES_MINUS_ENTRIES)
    elif IS_64BIT and fun == FUNC_INT:
        assert d.num_live_items < ((1 << 32) - MIN_INDEXES_MINUS_ENTRIES)
        return new_allocated > ((1 << 32) - MIN_INDEXES_MINUS_ENTRIES)
    return False

@jit.dont_look_inside
def ll_dict_remove_deleted_items(d):
    if d.num_live_items < len(d.entries) // 4:
        # At least 75% of the allocated entries are dead, so shrink the memory
        # allocated as well as doing a compaction.
        new_allocated = _overallocate_entries_len(d.num_live_items)
        newitems = lltype.malloc(lltype.typeOf(d).TO.entries.TO, new_allocated)
    else:
        newitems = d.entries
        # The loop below does a lot of writes into 'newitems'.  It's a better
        # idea to do a single gc_writebarrier rather than activating the
        # card-by-card logic (worth 11% in microbenchmarks).
        from rpython.rtyper.lltypesystem.lloperation import llop
        llop.gc_writebarrier(lltype.Void, newitems)
    #
    ENTRIES = lltype.typeOf(d).TO.entries.TO
    ENTRY = ENTRIES.OF
    isrc = 0
    idst = 0
    isrclimit = d.num_ever_used_items
    while isrc < isrclimit:
        if d.entries.valid(isrc):
            src = d.entries[isrc]
            dst = newitems[idst]
            dst.key = src.key
            dst.value = src.value
            if hasattr(ENTRY, 'f_hash'):
                dst.f_hash = src.f_hash
            if hasattr(ENTRY, 'f_valid'):
                assert src.f_valid
                dst.f_valid = True
            idst += 1
        isrc += 1
    assert d.num_live_items == idst
    d.num_ever_used_items = idst
    if ((ENTRIES.must_clear_key or ENTRIES.must_clear_value) and
            d.entries == newitems):
        # must clear the extra entries: they may contain valid pointers
        # which would create a temporary memory leak
        while idst < isrclimit:
            entry = newitems[idst]
            if ENTRIES.must_clear_key:
                entry.key = lltype.nullptr(ENTRY.key.TO)
            if ENTRIES.must_clear_value:
                entry.value = lltype.nullptr(ENTRY.value.TO)
            idst += 1
    else:
        d.entries = newitems

    ll_dict_reindex(d, _ll_len_of_d_indexes(d))


def ll_dict_delitem(d, key):
    ll_dict_delitem_with_hash(d, key, d.keyhash(key))

def ll_dict_delitem_with_hash(d, key, hash):
    index = d.lookup_function(d, key, hash, FLAG_LOOKUP)
    if index < 0:
        raise KeyError
    _ll_dict_del(d, hash, index)

def ll_dict_delitem_if_value_is(d, key, value):
    hash = d.keyhash(key)
    index = d.lookup_function(d, key, hash, FLAG_LOOKUP)
    if index < 0:
        return
    if d.entries[index].value != value:
        return
    _ll_dict_del(d, hash, index)

def _ll_dict_del_entry(d, index):
    d.entries.mark_deleted(index)
    d.num_live_items -= 1
    # clear the key and the value if they are GC pointers
    ENTRIES = lltype.typeOf(d.entries).TO
    ENTRY = ENTRIES.OF
    entry = d.entries[index]
    if ENTRIES.must_clear_key:
        entry.key = lltype.nullptr(ENTRY.key.TO)
    if ENTRIES.must_clear_value:
        entry.value = lltype.nullptr(ENTRY.value.TO)

@jit.look_inside_iff(lambda d, h, i: jit.isvirtual(d) and jit.isconstant(i))
def _ll_dict_del(d, hash, index):
    ll_call_delete_by_entry_index(d, hash, index, DELETED)
    _ll_dict_del_entry(d, index)

    if d.num_live_items == 0:
        # Dict is now empty.  Reset these fields.
        d.num_ever_used_items = 0
        d.lookup_function_no &= FUNC_MASK

    elif index == d.num_ever_used_items - 1:
        # The last element of the ordereddict has been deleted. Instead of
        # simply marking the item as dead, we can safely reuse it. Since it's
        # also possible that there are more dead items immediately behind the
        # last one, we reclaim all the dead items at the end of the ordereditem
        # at the same point.
        i = index
        while True:
            i -= 1
            assert i >= 0
            if d.entries.valid(i):    # must be at least one
                break
        d.num_ever_used_items = i + 1

    # If the dictionary is at least 87.5% dead items, then consider shrinking
    # it.
    if d.num_live_items + DICT_INITSIZE <= len(d.entries) / 8:
        ll_dict_resize(d)

def ll_dict_resize(d):
    # make a 'new_size' estimate and shrink it if there are many
    # deleted entry markers.  See CPython for why it is a good idea to
    # quadruple the dictionary size as long as it's not too big.
    # (Quadrupling comes from '(d.num_live_items + d.num_live_items + 1) * 2'
    # as long as num_live_items is not too large.)
    num_extra = min(d.num_live_items + 1, 30000)
    _ll_dict_resize_to(d, num_extra)
ll_dict_resize.oopspec = 'odict.resize(d)'

def _ll_dict_resize_to(d, num_extra):
    new_estimate = (d.num_live_items + num_extra) * 2
    new_size = DICT_INITSIZE
    while new_size <= new_estimate:
        new_size *= 2

    if new_size < _ll_len_of_d_indexes(d):
        ll_dict_remove_deleted_items(d)
    else:
        ll_dict_reindex(d, new_size)

def ll_ensure_indexes(d):
    num = d.lookup_function_no
    if num == FUNC_MUST_REINDEX:
        ll_dict_create_initial_index(d)
    else:
        ll_assert((num & FUNC_MASK) != FUNC_MUST_REINDEX,
                  "bad combination in lookup_function_no")

@jit.look_inside_iff(lambda d: jit.isvirtual(d))
def ll_dict_create_initial_index(d):
    """Create the initial index for a dictionary.  The common case is
    that 'd' is empty.  The uncommon case is that it is a prebuilt
    dictionary frozen by translation, in which case we must rehash all
    entries.  The common case must be seen by the JIT.
    """
    if d.num_live_items == 0:
        ll_malloc_indexes_and_choose_lookup(d, DICT_INITSIZE)
        d.resize_counter = DICT_INITSIZE * 2
    else:
        ll_dict_rehash_after_translation(d)

@jit.dont_look_inside
def ll_dict_rehash_after_translation(d):
    assert d.num_live_items == d.num_ever_used_items
    assert not d.indexes
    #
    # recompute all hashes.  Needed if they are stored in d.entries,
    # but do it anyway: otherwise, e.g. a string-keyed dictionary
    # won't have a fasthash on its strings if their hash is still
    # uncomputed.
    ENTRY = lltype.typeOf(d.entries).TO.OF
    for i in range(d.num_ever_used_items):
        assert d.entries.valid(i)
        d_entry = d.entries[i]
        h = d.keyhash(d_entry.key)
        if hasattr(ENTRY, 'f_hash'):
            d_entry.f_hash = h
        #else: purely for the side-effect it can have on d_entry.key
    #
    # Use the smallest acceptable size for ll_dict_reindex
    new_size = DICT_INITSIZE
    while new_size * 2 - d.num_live_items * 3 <= 0:
        new_size *= 2
    ll_dict_reindex(d, new_size)

def ll_dict_reindex(d, new_size):
    if bool(d.indexes) and _ll_len_of_d_indexes(d) == new_size:
        ll_clear_indexes(d, new_size)   # hack: we can reuse the same array
    else:
        ll_malloc_indexes_and_choose_lookup(d, new_size)
    d.resize_counter = new_size * 2 - d.num_live_items * 3
    ll_assert(d.resize_counter > 0, "reindex: resize_counter <= 0")
    ll_assert((d.lookup_function_no >> FUNC_SHIFT) == 0,
              "reindex: lookup_fun >> SHIFT")
    #
    entries = d.entries
    i = 0
    ibound = d.num_ever_used_items
    #
    # Write four loops, moving the check for the value of 'fun' out of
    # the loops.  A small speed-up over ll_call_insert_clean_function().
    fun = d.lookup_function_no     # == lookup_function_no & FUNC_MASK
    if fun == FUNC_BYTE:
        while i < ibound:
            if entries.valid(i):
                ll_dict_store_clean(d, entries.entry_hash(d, i), i, TYPE_BYTE)
            i += 1
    elif fun == FUNC_SHORT:
        while i < ibound:
            if entries.valid(i):
                ll_dict_store_clean(d, entries.entry_hash(d, i), i, TYPE_SHORT)
            i += 1
    elif IS_64BIT and fun == FUNC_INT:
        while i < ibound:
            if entries.valid(i):
                ll_dict_store_clean(d, entries.entry_hash(d, i), i, TYPE_INT)
            i += 1
    elif fun == FUNC_LONG:
        while i < ibound:
            if entries.valid(i):
                ll_dict_store_clean(d, entries.entry_hash(d, i), i, TYPE_LONG)
            i += 1
    else:
        assert False


# ------- a port of CPython's dictobject.c's lookdict implementation -------
PERTURB_SHIFT = 5

FREE = 0
DELETED = 1
VALID_OFFSET = 2
MIN_INDEXES_MINUS_ENTRIES = VALID_OFFSET + 1

FLAG_LOOKUP = 0
FLAG_STORE = 1

@specialize.memo()
def _ll_ptr_to_array_of(T):
    return lltype.Ptr(lltype.GcArray(T))

@jit.look_inside_iff(lambda d, key, hash, store_flag, T:
                     jit.isvirtual(d) and jit.isconstant(key))
@jit.oopspec('ordereddict.lookup(d, key, hash, store_flag, T)')
def ll_dict_lookup(d, key, hash, store_flag, T):
    INDEXES = _ll_ptr_to_array_of(T)
    entries = d.entries
    indexes = lltype.cast_opaque_ptr(INDEXES, d.indexes)
    mask = len(indexes) - 1
    i = r_uint(hash & mask)
    # do the first try before any looping
    ENTRIES = lltype.typeOf(entries).TO
    direct_compare = not hasattr(ENTRIES, 'no_direct_compare')
    index = rffi.cast(lltype.Signed, indexes[intmask(i)])
    if index >= VALID_OFFSET:
        checkingkey = entries[index - VALID_OFFSET].key
        if direct_compare and checkingkey == key:
            return index - VALID_OFFSET   # found the entry
        if d.keyeq is not None and entries.entry_hash(d, index - VALID_OFFSET) == hash:
            # correct hash, maybe the key is e.g. a different pointer to
            # an equal object
            found = d.keyeq(checkingkey, key)
            #llop.debug_print(lltype.Void, "comparing keys", ll_debugrepr(checkingkey), ll_debugrepr(key), found)
            if d.paranoia:
                if (entries != d.entries or lltype.cast_opaque_ptr(llmemory.GCREF, indexes) != d.indexes or
                    not entries.valid(index - VALID_OFFSET) or
                    entries[index - VALID_OFFSET].key != checkingkey):
                    # the compare did major nasty stuff to the dict: start over
                    return ll_dict_lookup(d, key, hash, store_flag, T)
            if found:
                return index - VALID_OFFSET
        deletedslot = -1
    elif index == DELETED:
        deletedslot = intmask(i)
    else:
        # pristine entry -- lookup failed
        if store_flag == FLAG_STORE:
            _ll_write_indexes(d, i, d.num_ever_used_items + VALID_OFFSET, T)
        return -1

    # In the loop, a deleted entry (everused and not valid) is by far
    # (factor of 100s) the least likely outcome, so test for that last.
    perturb = r_uint(hash)
    while 1:
        # compute the next index using unsigned arithmetic
        i = (i << 2) + i + perturb + 1
        i = i & mask
        index = rffi.cast(lltype.Signed, indexes[intmask(i)])
        if index == FREE:
            if store_flag == FLAG_STORE:
                if deletedslot == -1:
                    deletedslot = intmask(i)
                _ll_write_indexes(d, deletedslot, d.num_ever_used_items + VALID_OFFSET, T)
            return -1
        elif index >= VALID_OFFSET:
            checkingkey = entries[index - VALID_OFFSET].key
            if direct_compare and checkingkey == key:
                return index - VALID_OFFSET   # found the entry
            if d.keyeq is not None and entries.entry_hash(d, index - VALID_OFFSET) == hash:
                # correct hash, maybe the key is e.g. a different pointer to
                # an equal object
                found = d.keyeq(checkingkey, key)
                if d.paranoia:
                    if (entries != d.entries or lltype.cast_opaque_ptr(llmemory.GCREF, indexes) != d.indexes or
                        not entries.valid(index - VALID_OFFSET) or
                        entries[index - VALID_OFFSET].key != checkingkey):
                        # the compare did major nasty stuff to the dict: start over
                        return ll_dict_lookup(d, key, hash, store_flag, T)
                if found:
                    return index - VALID_OFFSET
        elif deletedslot == -1:
            deletedslot = intmask(i)
        perturb >>= PERTURB_SHIFT

def ll_dict_store_clean(d, hash, index, T):
    # a simplified version of ll_dict_lookup() which assumes that the
    # key is new, and the dictionary doesn't contain deleted entries.
    # It only finds the next free slot for the given hash.
    INDEXES = _ll_ptr_to_array_of(T)
    indexes = lltype.cast_opaque_ptr(INDEXES, d.indexes)
    mask = len(indexes) - 1
    i = r_uint(hash & mask)
    perturb = r_uint(hash)
    while rffi.cast(lltype.Signed, indexes[i]) != FREE:
        i = (i << 2) + i + perturb + 1
        i = i & mask
        perturb >>= PERTURB_SHIFT
    _ll_write_indexes(d, i, index + VALID_OFFSET, T)

# the following function is only called from _ll_dict_del, whose
# @jit.look_inside_iff condition should control when we get inside
# here with the jit
@jit.unroll_safe
def ll_dict_delete_by_entry_index(d, hash, locate_index, replace_with, T):
    # Another simplified version of ll_dict_lookup() which locates a
    # hashtable entry with the given 'index' stored in it, and deletes it.
    # This *should* be safe against evil user-level __eq__/__hash__
    # functions because the 'hash' argument here should be the one stored
    # into the directory, which is correct.
    INDEXES = _ll_ptr_to_array_of(T)
    indexes = lltype.cast_opaque_ptr(INDEXES, d.indexes)
    mask = len(indexes) - 1
    i = r_uint(hash & mask)
    perturb = r_uint(hash)
    locate_value = locate_index + VALID_OFFSET
    while rffi.cast(lltype.Signed, indexes[i]) != locate_value:
        assert rffi.cast(lltype.Signed, indexes[i]) != FREE
        i = (i << 2) + i + perturb + 1
        i = i & mask
        perturb >>= PERTURB_SHIFT
    _ll_write_indexes(d, i, replace_with, T)

# ____________________________________________________________
#
#  Irregular operations.

# Start the hashtable size at 16 rather than 8, as with rdict.py, because
# it is only an array of bytes
DICT_INITSIZE = 16


@specialize.memo()
def _ll_empty_array(DICT):
    """Memo function: cache a single prebuilt allocated empty array."""
    return DICT.entries.TO.allocate(0)

def ll_newdict(DICT):
    d = DICT.allocate()
    d.entries = _ll_empty_array(DICT)
    # Don't allocate an 'indexes' for empty dict.  It seems a typical
    # program contains tons of empty dicts, so this might be a memory win.
    d.num_live_items = 0
    d.num_ever_used_items = 0
    ll_no_initial_index(d)
    return d
OrderedDictRepr.ll_newdict = staticmethod(ll_newdict)

def ll_newdict_size(DICT, orig_length_estimate):
    length_estimate = (orig_length_estimate // 2) * 3
    n = DICT_INITSIZE
    while n < length_estimate:
        n *= 2
    d = DICT.allocate()
    d.entries = DICT.entries.TO.allocate(orig_length_estimate)
    ll_malloc_indexes_and_choose_lookup(d, n)
    d.num_live_items = 0
    d.num_ever_used_items = 0
    d.resize_counter = n * 2
    return d

# rpython.memory.lldict uses a dict based on Struct and Array
# instead of GcStruct and GcArray, which is done by using different
# 'allocate' and 'delete' adtmethod implementations than the ones below
def _ll_malloc_dict(DICT):
    return lltype.malloc(DICT)
def _ll_malloc_entries(ENTRIES, n):
    return lltype.malloc(ENTRIES, n, zero=True)
def _ll_free_entries(entries):
    pass


# ____________________________________________________________
#
#  Iteration.

def get_ll_dictiter(DICTPTR):
    return lltype.Ptr(lltype.GcStruct('dictiter',
                                      ('dict', DICTPTR),
                                      ('index', lltype.Signed)))

class DictIteratorRepr(AbstractDictIteratorRepr):

    def __init__(self, r_dict, variant="keys"):
        self.r_dict = r_dict
        self.variant = variant
        self.lowleveltype = get_ll_dictiter(r_dict.lowleveltype)
        if variant == 'reversed':
            self.ll_dictiter = ll_dictiter_reversed
            self._ll_dictnext = _ll_dictnext_reversed
        else:
            self.ll_dictiter = ll_dictiter
            self._ll_dictnext = _ll_dictnext


def ll_dictiter(ITERPTR, d):
    iter = lltype.malloc(ITERPTR.TO)
    iter.dict = d
    # initialize the index with usually 0, but occasionally a larger value
    iter.index = d.lookup_function_no >> FUNC_SHIFT
    return iter

@jit.look_inside_iff(lambda iter: jit.isvirtual(iter)
                     and (iter.dict is None or
                          jit.isvirtual(iter.dict)))
@jit.oopspec("odictiter.next(iter)")
def _ll_dictnext(iter):
    dict = iter.dict
    if dict:
        entries = dict.entries
        index = iter.index
        assert index >= 0
        entries_len = dict.num_ever_used_items
        while index < entries_len:
            nextindex = index + 1
            if entries.valid(index):
                iter.index = nextindex
                return index
            else:
                # In case of repeated iteration over the start of
                # a dict where the items get removed, like
                # collections.OrderedDict.popitem(last=False),
                # the hack below will increase the value stored in
                # the high bits of lookup_function_no and so the
                # next iteration will start at a higher value.
                # We should carefully reset these high bits to zero
                # as soon as we do something like ll_dict_reindex().
                if index == (dict.lookup_function_no >> FUNC_SHIFT):
                    dict.lookup_function_no += (1 << FUNC_SHIFT)
                # note that we can't have modified a FUNC_MUST_REINDEX
                # dict here because such dicts have no invalid entries
                ll_assert((dict.lookup_function_no & FUNC_MASK) !=
                      FUNC_MUST_REINDEX, "bad combination in _ll_dictnext")
            index = nextindex
        # clear the reference to the dict and prevent restarts
        iter.dict = lltype.nullptr(lltype.typeOf(iter).TO.dict.TO)
    raise StopIteration

def ll_dictiter_reversed(ITERPTR, d):
    iter = lltype.malloc(ITERPTR.TO)
    iter.dict = d
    iter.index = d.num_ever_used_items
    return iter

def _ll_dictnext_reversed(iter):
    dict = iter.dict
    if dict:
        entries = dict.entries
        index = iter.index - 1
        while index >= 0:
            if entries.valid(index):
                iter.index = index
                return index
            index = index - 1
        # clear the reference to the dict and prevent restarts
        iter.dict = lltype.nullptr(lltype.typeOf(iter).TO.dict.TO)
    raise StopIteration

# _____________________________________________________________
# methods

def ll_dict_get(dict, key, default):
    index = dict.lookup_function(dict, key, dict.keyhash(key), FLAG_LOOKUP)
    if index < 0:
        return default
    else:
        return dict.entries[index].value

def ll_dict_setdefault(dict, key, default):
    hash = dict.keyhash(key)
    index = dict.lookup_function(dict, key, hash, FLAG_STORE)
    if index < 0:
        _ll_dict_setitem_lookup_done(dict, key, default, hash, -1)
        return default
    else:
        return dict.entries[index].value

@jit.look_inside_iff(lambda dict: jit.isvirtual(dict))
def ll_dict_copy(dict):
    DICT = lltype.typeOf(dict).TO
    if dict.num_live_items == 0:
        return ll_newdict(DICT)

    # the only dicts without index are either: empty, or prebuilt. we checked
    # for empty in the if above. therefore, we never have to reindex while
    # jitting, because dict is virtual and thus not prebuilt
    if not jit.we_are_jitted():
        ll_ensure_indexes(dict)
    else:
        num = dict.lookup_function_no
        assert (num & FUNC_MASK) != FUNC_MUST_REINDEX

    newdict = DICT.allocate()
    newdict.entries = DICT.entries.TO.allocate(len(dict.entries))

    newdict.num_live_items = dict.num_live_items
    newdict.num_ever_used_items = dict.num_ever_used_items
    if hasattr(DICT, 'fnkeyeq'):
        newdict.fnkeyeq = dict.fnkeyeq
    if hasattr(DICT, 'fnkeyhash'):
        newdict.fnkeyhash = dict.fnkeyhash

    rgc.ll_arraycopy(dict.entries, newdict.entries, 0, 0,
                     newdict.num_ever_used_items)

    fun = dict.lookup_function_no & FUNC_MASK
    if fun == FUNC_BYTE:
        indexes = lltype.cast_opaque_ptr(DICTINDEX_BYTE, dict.indexes)
        n = len(indexes)
        newindexes = lltype.malloc(DICTINDEX_BYTE.TO, n, zero=True)
        rgc.ll_arraycopy(indexes, newindexes, 0, 0, n)
        newdict.indexes = lltype.cast_opaque_ptr(llmemory.GCREF, newindexes)
        newdict.lookup_function_no = FUNC_BYTE
    elif fun == FUNC_SHORT:
        indexes = lltype.cast_opaque_ptr(DICTINDEX_SHORT, dict.indexes)
        n = len(indexes)
        newindexes = lltype.malloc(DICTINDEX_SHORT.TO, n, zero=True)
        rgc.ll_arraycopy(indexes, newindexes, 0, 0, n)
        newdict.indexes = lltype.cast_opaque_ptr(llmemory.GCREF, newindexes)
        newdict.lookup_function_no = FUNC_SHORT
    elif IS_64BIT and fun == FUNC_INT:
        indexes = lltype.cast_opaque_ptr(DICTINDEX_INT, dict.indexes)
        n = len(indexes)
        newindexes = lltype.malloc(DICTINDEX_INT.TO, n, zero=True)
        rgc.ll_arraycopy(indexes, newindexes, 0, 0, n)
        newdict.indexes = lltype.cast_opaque_ptr(llmemory.GCREF, newindexes)
        newdict.lookup_function_no = FUNC_INT
    else:
        indexes = lltype.cast_opaque_ptr(DICTINDEX_LONG, dict.indexes)
        n = len(indexes)
        newindexes = lltype.malloc(DICTINDEX_LONG.TO, n, zero=True)
        rgc.ll_arraycopy(indexes, newindexes, 0, 0, n)
        newdict.indexes = lltype.cast_opaque_ptr(llmemory.GCREF, newindexes)
        newdict.lookup_function_no = FUNC_LONG
    newdict.resize_counter = dict.resize_counter
    return newdict

def ll_dict_clear(d):
    if d.num_ever_used_items == 0:
        return
    DICT = lltype.typeOf(d).TO
    old_entries = d.entries
    d.entries = _ll_empty_array(DICT)
    # note: we can't remove the index here, because it is possible that
    # crazy Python code calls d.clear() from the method __eq__() called
    # from ll_dict_lookup(d).  Instead, stick to the rule that once a
    # dictionary has got an index, it will always have one.
    ll_malloc_indexes_and_choose_lookup(d, DICT_INITSIZE)
    d.num_live_items = 0
    d.num_ever_used_items = 0
    d.resize_counter = DICT_INITSIZE * 2
    # old_entries.delete() XXX
ll_dict_clear.oopspec = 'odict.clear(d)'

@jit.look_inside_iff(lambda dic1, dic2: jit.isvirtual(dic1) and jit.isvirtual(dic2))
def ll_dict_update(dic1, dic2):
    if dic1 == dic2:
        return
    ll_ensure_indexes(dic2)    # needed for entries.entry_hash() below
    ll_prepare_dict_update(dic1, dic2.num_live_items)
    i = 0
    while i < dic2.num_ever_used_items:
        entries = dic2.entries
        if entries.valid(i):
            entry = entries[i]
            hash = entries.entry_hash(dic2, i)
            key = entry.key
            value = entry.value
            index = dic1.lookup_function(dic1, key, hash, FLAG_STORE)
            _ll_dict_setitem_lookup_done(dic1, key, value, hash, index)
        i += 1

def ll_prepare_dict_update(d, num_extra):
    # Prescale 'd' for 'num_extra' items, assuming that most items don't
    # collide.  If this assumption is false, 'd' becomes too large by at
    # most 'num_extra'.  The logic is based on:
    #      (d.resize_counter - 1) // 3 = room left in d
    #  so, if num_extra == 1, we need d.resize_counter > 3
    #      if num_extra == 2, we need d.resize_counter > 6  etc.
    # Note however a further hack: if num_extra <= d.num_live_items,
    # we avoid calling _ll_dict_resize_to here.  This is to handle
    # the case where dict.update() actually has a lot of collisions.
    # If num_extra is much greater than d.num_live_items the conditional_call
    # will trigger anyway, which is really the goal.
    ll_ensure_indexes(d)
    x = num_extra - d.num_live_items
    jit.conditional_call(d.resize_counter <= x * 3,
                         _ll_dict_resize_to, d, num_extra)

# this is an implementation of keys(), values() and items()
# in a single function.
# note that by specialization on func, three different
# and very efficient functions are created.

def recast(P, v):
    # XXX this function is a terrible hack
    if P is llmemory.GCREF:
        return lltype.cast_opaque_ptr(llmemory.GCREF, v)
    if lltype.typeOf(v) is llmemory.GCREF:
        return lltype.cast_opaque_ptr(P, v)
    if isinstance(P, lltype.Ptr):
        return lltype.cast_pointer(P, v)
    else:
        return v

def _make_ll_keys_values_items(kind):
    def ll_kvi(LIST, dic, EXTERNAL_ELEM=None):
        res = LIST.ll_newlist(dic.num_live_items)
        entries = dic.entries
        dlen = dic.num_ever_used_items
        items = res.ll_items()
        ELEM = lltype.typeOf(items).TO.OF
        i = 0
        p = 0
        while i < dlen:
            if entries.valid(i):
                if ELEM is not lltype.Void:
                    entry = entries[i]
                    if kind == 'items':
                        r = lltype.malloc(EXTERNAL_ELEM.TO)
                        r.item0 = recast(EXTERNAL_ELEM.TO.item0, entry.key)
                        r.item1 = recast(EXTERNAL_ELEM.TO.item1, entry.value)
                        items[p] = recast(ELEM, r)
                    elif kind == 'keys':
                        items[p] = recast(ELEM, entry.key)
                    elif kind == 'values':
                        items[p] = recast(ELEM, entry.value)
                p += 1
            i += 1
        assert p == res.ll_length()
        return res
    if kind != "items":
        ll_kvi.oopspec = 'odict.%s(dic)' % kind
    return ll_kvi

ll_dict_keys   = _make_ll_keys_values_items('keys')
ll_dict_values = _make_ll_keys_values_items('values')
ll_dict_items  = _make_ll_keys_values_items('items')

def ll_dict_contains(d, key):
    return ll_dict_contains_with_hash(d, key, d.keyhash(key))

def ll_dict_contains_with_hash(d, key, hash):
    i = d.lookup_function(d, key, hash, FLAG_LOOKUP)
    return i >= 0

def _ll_getnextitem(dic):
    if dic.num_live_items == 0:
        raise KeyError

    ll_ensure_indexes(dic)
    entries = dic.entries

    # find the last entry.  It's unclear if the loop below is still
    # needed nowadays, because 'num_ever_used_items - 1' should always
    # point to the last active item (we decrease it as needed in
    # _ll_dict_del).  Better safe than sorry.
    while True:
        i = dic.num_ever_used_items - 1
        if entries.valid(i):
            break
        dic.num_ever_used_items -= 1

    return i

def ll_dict_popitem(ELEM, dic):
    i = _ll_getnextitem(dic)
    entry = dic.entries[i]
    r = lltype.malloc(ELEM.TO)
    r.item0 = recast(ELEM.TO.item0, entry.key)
    r.item1 = recast(ELEM.TO.item1, entry.value)
    _ll_dict_del(dic, dic.entries.entry_hash(dic, i), i)
    return r

def ll_dict_pop(dic, key):
    hash = dic.keyhash(key)
    index = dic.lookup_function(dic, key, hash, FLAG_LOOKUP)
    if index < 0:
        raise KeyError
    value = dic.entries[index].value
    _ll_dict_del(dic, hash, index)
    return value

def ll_dict_pop_default(dic, key, dfl):
    hash = dic.keyhash(key)
    index = dic.lookup_function(dic, key, hash, FLAG_LOOKUP)
    if index < 0:
        return dfl
    value = dic.entries[index].value
    _ll_dict_del(dic, hash, index)
    return value

def ll_dict_move_to_end(d, key, last):
    if last:
        ll_dict_move_to_last(d, key)
    else:
        ll_dict_move_to_first(d, key)

def ll_dict_move_to_last(d, key):
    hash = d.keyhash(key)
    old_index = d.lookup_function(d, key, hash, FLAG_LOOKUP)
    if old_index < 0:
        raise KeyError

    if old_index == d.num_ever_used_items - 1:
        return

    # remove the entry at the old position
    old_entry = d.entries[old_index]
    key = old_entry.key
    value = old_entry.value
    _ll_dict_del_entry(d, old_index)

    # note a corner case: it is possible that 'replace_with' is just too
    # large to fit in the type T used so far for the index.  But in that
    # case, the list 'd.entries' is full, and the following call to
    # _ll_dict_setitem_lookup_done() will necessarily reindex the dict.
    # So in that case, this value of 'replace_with' should be ignored.
    ll_call_delete_by_entry_index(d, hash, old_index,
            replace_with = VALID_OFFSET + d.num_ever_used_items)
    _ll_dict_setitem_lookup_done(d, key, value, hash, -1)

def ll_dict_move_to_first(d, key):
    # In this function, we might do a bit more than the strict minimum
    # of walks over parts of the array, trying to keep the code at least
    # semi-reasonable, while the goal is still amortized constant-time
    # over many calls.

    # Call ll_dict_remove_deleted_items() first if there are too many
    # deleted items.  Not a perfect solution, because lookup_function()
    # might do random things with the dict and create many new deleted
    # items.  Still, should be fine, because nothing crucially depends
    # on this: the goal is to avoid the dictionary's list growing
    # forever.
    if d.num_live_items < len(d.entries) // 2 - 16:
        ll_dict_remove_deleted_items(d)

    hash = d.keyhash(key)
    old_index = d.lookup_function(d, key, hash, FLAG_LOOKUP)
    if old_index <= 0:
        if old_index < 0:
            raise KeyError
        else:
            # already at the beginning
            return

    # the goal of the following is to set 'idst' to the number of
    # deleted entries at the beginning, ensuring 'idst > 0'
    must_reindex = False
    if d.entries.valid(0):
        # the first entry is valid, so we need to make room before.
        # to get the complexity right if move_to_first is called a lot, we want
        # to overallocate
        must_reindex = True
        new_allocated = _overallocate_entries_len(d.num_ever_used_items)
        if not _ll_dict_entries_size_too_big(d, new_allocated):
            # use 75% of the new space at the beginning
            idst = ((new_allocated - d.num_ever_used_items) * 3) // 4
            ll_assert(idst > 0, "overallocate did not do enough")
            newitems = lltype.malloc(lltype.typeOf(d).TO.entries.TO, new_allocated)
            rgc.ll_arraycopy(d.entries, newitems, 0, idst, d.num_ever_used_items)
            d.entries = newitems
            i = 0
            while i < idst:
                d.entries.mark_deleted(i)
                i += 1
            d.num_ever_used_items += idst
            old_index += idst
            idst -= 1
            # XXX we could set lookup_function_no with idst here
        else:
            # see comment in ll_dict_grow: we have lots of deleted entries
            # towards the end of the dict, but not enough in the front. rescue
            # this by compacting the live items towards the end
            ll_assert(d.num_live_items < d.num_ever_used_items, "deleted entries exist, but num_live_items == num_ever_used_items?!")
            old_index = _ll_dict_move_to_first_shift_items(d, old_index)
            idst = d.lookup_function_no >> FUNC_SHIFT
            idst -= 1
    else:
        idst = d.lookup_function_no >> FUNC_SHIFT
        # All entries in range(0, idst) are deleted.  Check if more are
        while not d.entries.valid(idst):
            idst += 1
        if idst == old_index:
            d.lookup_function_no = ((d.lookup_function_no & FUNC_MASK) |
                                    (old_index << FUNC_SHIFT))
            return
        idst -= 1
        d.lookup_function_no = ((d.lookup_function_no & FUNC_MASK) |
                                (idst << FUNC_SHIFT))

    # remove the entry at the old position
    ll_assert(d.entries.valid(old_index),
              "ll_dict_move_to_first: lost old_index")
    ENTRY = lltype.typeOf(d.entries).TO.OF
    old_entry = d.entries[old_index]
    key = old_entry.key
    value = old_entry.value
    if hasattr(ENTRY, 'f_hash'):
        ll_assert(old_entry.f_hash == hash,
                  "ll_dict_move_to_first: bad hash")
    _ll_dict_del_entry(d, old_index)

    # put the entry at its new position
    ll_assert(not d.entries.valid(idst),
              "ll_dict_move_to_first: overwriting idst")
    new_entry = d.entries[idst]
    new_entry.key = key
    new_entry.value = value
    if hasattr(ENTRY, 'f_hash'):
        new_entry.f_hash = hash
    if hasattr(ENTRY, 'f_valid'):
        new_entry.f_valid = True
    d.num_live_items += 1

    # fix the index
    if must_reindex:
        ll_dict_reindex(d, _ll_len_of_d_indexes(d))
    else:
        ll_call_delete_by_entry_index(d, hash, old_index,
                replace_with = VALID_OFFSET + idst)

def _ll_dict_move_to_first_shift_items(d, old_index):
    from rpython.rtyper.lltypesystem.lloperation import llop
    # situation: lots of deleted keys somewhere in the dict (not at the front).
    # compact the dict towards the end. num_ever_used_items does not change,
    # the items are just compacted (leaving free space at the end free).
    # returns the new index of the entry that used to be at old_index
    entries = d.entries
    llop.gc_writebarrier(lltype.Void, entries)
    #
    ENTRIES = lltype.typeOf(d).TO.entries.TO
    ENTRY = ENTRIES.OF
    idst = isrc = d.num_ever_used_items - 1
    new_index = -1
    while isrc >= 0:
        if entries.valid(isrc):
            if isrc == old_index:
                new_index = idst
            src = entries[isrc]
            dst = entries[idst]
            dst.key = src.key
            dst.value = src.value
            if hasattr(ENTRY, 'f_hash'):
                dst.f_hash = src.f_hash
            if hasattr(ENTRY, 'f_valid'):
                assert src.f_valid
                dst.f_valid = True
            idst -= 1
        isrc -= 1
    idst += 1 # range(0, idst) is empty now
    assert d.num_ever_used_items - d.num_live_items == idst
    assert new_index >= 0
    # set the first idst items to be invalid
    i = 0
    while i < idst:
        d.entries.mark_deleted(i)
        i += 1
    d.lookup_function_no = ((d.lookup_function_no & FUNC_MASK) |
                            (idst << FUNC_SHIFT))
    return new_index



"""Support for Windows."""
from __future__ import print_function

import py, os, sys, re, shutil

from rpython.translator.platform import CompilationError
from rpython.translator.platform import log, _run_subprocess
from rpython.translator.platform import Platform, posix

import rpython
rpydir = str(py.path.local(rpython.__file__).join('..'))

def _get_compiler_type(cc, x64_flag):
    if not cc:
        cc = os.environ.get('CC','')
    if not cc:
        return MsvcPlatform(x64=x64_flag)
    elif cc.startswith('mingw') or cc == 'gcc':
        return MingwPlatform(cc)
    return MsvcPlatform(cc=cc, x64=x64_flag)

def _get_vcver0():
    # try to get the compiler which served to compile python
    msc_pos = sys.version.find('MSC v.')
    if msc_pos != -1:
        msc_ver = int(sys.version[msc_pos+6:msc_pos+10])
        # 1500 -> 90, 1900 -> 140
        vsver = (msc_ver / 10) - 60
        return vsver
    return None

def Windows(cc=None):
    return _get_compiler_type(cc, False)

def Windows_x64(cc=None, ver0=None):
    #raise Exception("Win64 is not supported.  You must either build for Win32"
    #                " or contribute the missing support in PyPy.")
    return _get_compiler_type(cc, True)

def _find_vcvarsall(version, x64flag):
    import rpython.tool.setuptools_msvc as msvc
    if x64flag:
        arch = 'x64'
    else:
        arch = 'x86'
    if version >= 140:
        return msvc.msvc14_get_vc_env(arch)
    else:
        return msvc.msvc9_query_vcvarsall(version / 10.0, arch)

def _get_msvc_env(vsver, x64flag):
    vcdict = None
    toolsdir = None
    try:
        toolsdir = os.environ['VS%sCOMNTOOLS' % vsver]
    except KeyError:
        # use setuptools from python3 to find tools
        try:
            vcdict = _find_vcvarsall(vsver, x64flag)
        except ImportError as e:
            if 'setuptools' in str(e):
                log.error('is setuptools installed (perhaps try %s -mensurepip)?' % sys.executable)
            log.error('looking for compiler %s raised exception "%s' % (vsver, str(e)))
        except Exception as e:
            log.error('looking for compiler %s raised exception "%s' % (vsver, str(e)))
            return None
    else:
        if x64flag:
            vsinstalldir = os.path.abspath(os.path.join(toolsdir, '..', '..'))
            vcinstalldir = os.path.join(vsinstalldir, 'VC')
            vcbindir = os.path.join(vcinstalldir, 'BIN')
            vcvars = os.path.join(vcbindir, 'amd64', 'vcvarsamd64.bat')
        else:
            vcvars = os.path.join(toolsdir, 'vsvars32.bat')
            if not os.path.exists(vcvars):
                # even msdn does not know which to run
                # see https://msdn.microsoft.com/en-us/library/1700bbwd(v=vs.90).aspx
                # which names both
                vcvars = os.path.join(toolsdir, 'vcvars32.bat')

        import subprocess
        try:
            popen = subprocess.Popen('"%s" & set' % (vcvars,),
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)

            stdout, stderr = popen.communicate()
            if popen.wait() != 0 or stdout[:5].lower() == 'error':
                log.msg('Running "%s" errored: \n\nstdout:\n%s\n\nstderr:\n%s' % (
                    vcvars, stdout.split()[0], stderr))
                return None
            else:
                log.msg('Running "%s" succeeded' %(vcvars,))
        except Exception as e:
            log.msg('Running "%s" failed: "%s"' % (vcvars, str(e)))
            return None

        stdout = stdout.replace("\r\n", "\n")
        vcdict = {}
        for line in stdout.split("\n"):
            if '=' not in line:
                continue
            key, value = line.split('=', 1)
            vcdict[key] = value
    env = {}
    for key, value in vcdict.items():
        if key.upper() in ['PATH', 'INCLUDE', 'LIB']:
            if sys.version_info[0] < 3:
                env[key.upper()] = value.encode('utf-8')
            else:
                env[key.upper()] = value
    if 'PATH' not in env:
        log.msg('Did not find "PATH" in stdout\n%s' %(stdout))
    if not _find_executable('mt.exe', env['PATH']):
        # For some reason the sdk bin path is missing?
        # put it together from some other env variables that happened to exist
        # on the buildbot where this occurred
        if 'WindowsSDKVersion' in vcdict and 'WindowsSdkDir' in vcdict:
            binpath = vcdict['WindowsSdkDir'] + '\\bin\\' + vcdict['WindowsSDKVersion'] + 'x86'
            env['PATH'] += ';' + binpath
        if not _find_executable('mt.exe', env['PATH']):
            log.msg('Could not find mt.exe on path=%s' % env['PATH'])
            log.msg('Running vsver %s set this env' % vsver)
            for key, value in vcdict.items():
                log.msg('%s=%s' %(key, value))
    log.msg("Updated environment with vsver %d, using x64 %s" % (vsver, x64flag,))
    return env

def find_msvc_env(x64flag=False, ver0=None):
    vcvers = [160, 150, 141, 140, 100, 90]
    if ver0 in vcvers:
        vcvers.insert(0, ver0)
    errs = []
    for vsver in vcvers:
        env = _get_msvc_env(vsver, x64flag)
        if env is not None:
            return env, vsver
    log.error("Could not find a Microsoft Compiler")
    # Assume that the compiler is already part of the environment

# copied from distutils.spawn
def _find_executable(executable, path=None):
    """Tries to find 'executable' in the directories listed in 'path'.

    A string listing directories separated by 'os.pathsep'; defaults to
    os.environ['PATH'].  Returns the complete filename or None if not found.
    """
    if path is None:
        path = os.environ['PATH']
    paths = path.split(os.pathsep)

    for ext in '.exe', '':
        newexe = executable + ext

        if os.path.isfile(newexe):
            return newexe
        else:
            for p in paths:
                f = os.path.join(p, newexe)
                if os.path.isfile(f):
                    # the file exists, we have a shot at spawn working
                    return f
    return None

class MsvcPlatform(Platform):
    name = "msvc"
    so_ext = 'dll'
    exe_ext = 'exe'

    relevant_environ = ('PATH', 'INCLUDE', 'LIB')

    cc = 'cl.exe'
    link = 'link.exe'
    make = 'nmake'
    if _find_executable('jom.exe'):
        make = 'jom.exe'

    cflags = ('/MD', '/O2', '/FS', '/Zi')
    # allow >2GB address space, set stack to 3MB (1MB is too small)
    link_flags = ('/nologo', '/debug','/LARGEADDRESSAWARE',
                  '/STACK:3145728', '/MANIFEST:EMBED')
    standalone_only = ()
    shared_only = ()
    environ = None

    def __init__(self, cc=None, x64=False, ver0=None):
        self.x64 = x64
        patch_os_env(self.externals)
        self.c_environ = os.environ.copy()
        if cc is None:
            msvc_compiler_environ, self.vsver = find_msvc_env(x64, ver0=ver0)
            Platform.__init__(self, 'cl.exe')
            if msvc_compiler_environ:
                self.c_environ.update(msvc_compiler_environ)
                self.version = "MSVC %s" % str(self.vsver)
                if self.vsver > 90:
                    tag = '14x'
                else:
                    tag = '%d' % self.vsver
                if x64:
                    self.externals_branch = 'win64_%s' % tag
                else:
                    self.externals_branch = 'win32_%s' % tag
        else:
            self.cc = cc

        # Try to find a masm assembler
        # Dilemma: raise now or later if masm is not found. Postponing the
        # exception means we can use a fake compiler for testing on linux
        # but may mean cryptic error messages and wasted build time.
        masm32 = "'Could not find ml.exe'"
        masm64 = "'Could not find ml.exe'"
        try:
            returncode, stdout, stderr = _run_subprocess(
                'ml.exe' if not x64 else 'ml64.exe', [], env=self.c_environ)
            r = re.search('Macro Assembler', stderr)
        except (EnvironmentError, OSError):
            r = None
        if r is not None:
            masm32 = 'ml.exe'
            masm64 = 'ml64.exe'
        else:
            if os.path.exists('c:/masm64/bin/masm64.exe'):
                masm64 = 'c:/masm64/bin/masm64.exe'
            if os.path.exists('c:/masm32/bin/masm32.exe'):
                masm64 = 'c:/masm32/bin/masm32.exe'
        if x64:
            self.masm = masm64
        else:
            self.masm = masm32

        # Install debug options only when interpreter is in debug mode
        if sys.executable.lower().endswith('_d.exe'):
            self.cflags = ['/MDd', '/Z7', '/Od']

            # Increase stack size, for the linker and the stack check code.
            stack_size = 8 << 20  # 8 Mb
            self.link_flags = self.link_flags + ('/STACK:%d' % stack_size,)
            # The following symbol is used in c/src/stack.h
            self.cflags.append('/DMAX_STACK_SIZE=%d' % (stack_size - 1024))

    def _includedirs(self, include_dirs):
        return ['/I%s' % (idir,) for idir in include_dirs]

    def _libs(self, libraries):
        libs = []
        for lib in libraries:
            lib = str(lib)
            if lib.endswith('.dll'):
                lib = lib[:-4]
            libs.append('%s.lib' % (lib,))
        libs.append("kernel32.lib")
        return libs

    def _libdirs(self, library_dirs):
        return ['/LIBPATH:%s' % (ldir,) for ldir in library_dirs]

    def _linkfiles(self, link_files):
        return list(link_files)

    def _args_for_shared(self, args, **kwds):
        return ['/dll'] + args

    def check___thread(self):
        # __declspec(thread) does not seem to work when using assembler.
        # Returning False will cause the program to use TlsAlloc functions.
        # see src/thread_win7.h
        return False

    def _link_args_from_eci(self, eci, standalone):
        # Windows needs to resolve all symbols even for DLLs
        return super(MsvcPlatform, self)._link_args_from_eci(eci, standalone=True)

    def _compile_c_file(self, cc, cfile, compile_args):
        oname = self._make_o_file(cfile, ext='obj')
        # notabene: (tismer)
        # This function may be called for .c but also .asm files.
        # The c compiler accepts any order of arguments, while
        # the assembler still has the old behavior that all options
        # must come first, and after the file name all options are ignored.
        # So please be careful with the order of parameters! ;-)
        pdb_dir = oname.dirname
        if pdb_dir:
                compile_args = compile_args + ['/Fd%s\\' % (pdb_dir,)]
        args = ['/nologo', '/c'] + compile_args + ['/Fo%s' % (oname,), str(cfile)]
        self._execute_c_compiler(cc, args, oname)
        return oname

    def _link(self, cc, ofiles, link_args, standalone, exe_name):
        args = ['/nologo'] + [str(ofile) for ofile in ofiles] + link_args
        args += ['/out:%s' % (exe_name,), '/incremental:no']
        if not standalone:
            args = self._args_for_shared(args)

        # Tell the linker to embed a manifest with the default
        # UAC level asInvoker (Visual Studio 2008 +)
        args += ["/MANIFEST:EMBED"]

        self._execute_c_compiler(self.link, args, exe_name)

        return exe_name

    def _handle_error(self, returncode, stdout, stderr, outname):
        if returncode != 0:
            # Microsoft compilers write compilation errors to stdout
            stderr = stdout + stderr
            errorfile = outname.new(ext='errors')
            errorfile.write(stderr, mode='wb')
            if self.log_errors:
                stderrlines = stderr.splitlines()
                for line in stderrlines:
                    log.Error(line)
                # ^^^ don't use ERROR, because it might actually be fine.
                # Also, ERROR confuses lib-python/conftest.py.
            raise CompilationError(stdout, stderr)


    def gen_makefile(self, cfiles, eci, exe_name=None, path=None,
                     shared=False, headers_to_precompile=[],
                     no_precompile_cfiles = [], profopt=False, config=None):
        cfiles = self._all_cfiles(cfiles, eci)

        if path is None:
            path = cfiles[0].dirpath()

        rpypath = py.path.local(rpydir)
        m = NMakefile(path)

        if exe_name is None:
            exe_name = cfiles[0].new(ext='')
        if shared:
            so_name = exe_name.new(purebasename='lib' + exe_name.basename,
                                   ext=self.so_ext)
            wtarget_name = exe_name.new(purebasename=exe_name.basename + 'w',
                                   ext=self.exe_ext)
            target_name = so_name.basename
            m.so_name = path.join(target_name)
            m.wtarget_name = path.join(wtarget_name.basename)
            m.exe_name = path.join(exe_name.basename + '.' + self.exe_ext)
        else:
            target_name = exe_name.basename + '.' + self.exe_ext
            wtarget_name = exe_name.basename + 'w.' + self.exe_ext
            m.exe_name = path.join(target_name)
            m.wtarget_name = path.join(wtarget_name)

        m.eci = eci

        linkflags = list(self.link_flags)
        if shared:
            linkflags = self._args_for_shared(linkflags)
        linkflags += self._exportsymbols_link_flags()
        # Make sure different functions end up at different addresses!
        # This is required for the JIT.
        linkflags.append('/opt:noicf')

        def rpyrel(fpath):
            rel = py.path.local(fpath).relto(rpypath)
            if rel:
                return os.path.join('$(RPYDIR)', rel)
            else:
                return fpath

        rel_cfiles = [m.pathrel(cfile) for cfile in cfiles]
        rel_ofiles = [rel_cfile[:rel_cfile.rfind('.')]+'.obj' for rel_cfile in rel_cfiles]
        m.cfiles = rel_cfiles

        rel_includedirs = [rpyrel(incldir) for incldir in
                           self.preprocess_include_dirs(eci.include_dirs)]
        rel_libdirs = [rpyrel(libdir) for libdir in
                       self.preprocess_library_dirs(eci.library_dirs)]

        m.comment('automatically generated makefile')
        definitions = [
            ('RPYDIR', '"%s"' % rpydir),
            ('TARGET', target_name),
            ('DEFAULT_TARGET', m.exe_name.basename),
            ('SOURCES', rel_cfiles),
            ('OBJECTS', rel_ofiles),
            ('LIBS', self._libs(eci.libraries)),
            ('LIBDIRS', self._libdirs(rel_libdirs)),
            ('INCLUDEDIRS', self._includedirs(rel_includedirs)),
            ('CFLAGS', self.cflags),
            ('CFLAGSEXTRA', list(eci.compile_extra)),
            ('LDFLAGS', linkflags),
            ('LDFLAGSEXTRA', list(eci.link_extra)),
            ('CC', self.cc),
            ('CC_LINK', self.link),
            ('LINKFILES', eci.link_files),
            ('MASM', self.masm),
            ('MAKE', 'nmake.exe'),
            ('_WIN32', '1'),
            ]
        if shared:
            definitions.insert(0, ('WTARGET', wtarget_name.basename))
        if self.x64:
            definitions.append(('_WIN64', '1'))

        if shared and self.make == 'jom.exe':
            # Add `.SYNC` for jom.exe and get it to create
            # main.c, wmain.c in a separate step before trying to compile them
            rules = [('all',
                      'main.c wmain.c .SYNC $(DEFAULT_TARGET) .SYNC $(WTARGET)',
                      []),
                    ]
        else:
            rules = [('all', '$(DEFAULT_TARGET) $(WTARGET)', []),]
        rules += [
            ('.asm.obj', '', '$(MASM) /nologo /Fo$@ /c $< $(INCLUDEDIRS)'),
            ]

        if len(headers_to_precompile)>0:
            if shared:
                no_precompile_cfiles += [m.makefile_dir / 'main.c',
                                         m.makefile_dir / 'wmain.c']
            stdafx_h = path.join('stdafx.h')
            txt  = '#ifndef PYPY_STDAFX_H\n'
            txt += '#define PYPY_STDAFX_H\n'
            txt += '\n'.join(['#include "' + m.pathrel(c) + '"' for c in headers_to_precompile])
            txt += '\n#endif\n'
            stdafx_h.write(txt)
            stdafx_c = path.join('stdafx.c')
            stdafx_c.write('#include "stdafx.h"\n')
            definitions.append(('CREATE_PCH', '/Ycstdafx.h /Fpstdafx.pch /FIstdafx.h'))
            definitions.append(('USE_PCH', '/Yustdafx.h /Fpstdafx.pch /FIstdafx.h'))
            rules.append(('$(OBJECTS)', 'stdafx.pch', []))
            rules.append(('stdafx.pch', 'stdafx.h',
               '$(CC) stdafx.c /c /nologo $(CFLAGS) $(CFLAGSEXTRA) '
               '$(CREATE_PCH) $(INCLUDEDIRS)'))
            rules.append(('.c.obj', '',
                    '$(CC) /nologo $(CFLAGS) $(CFLAGSEXTRA) $(USE_PCH) '
                    '/Fo$@ /c $< $(INCLUDEDIRS)'))
            #Do not use precompiled headers for some files
            #rules.append((r'{..\module_cache}.c{..\module_cache}.obj', '',
            #        '$(CC) /nologo $(CFLAGS) $(CFLAGSEXTRA) /Fo$@ /c $< $(INCLUDEDIRS)'))
            # nmake cannot handle wildcard target specifications, so we must
            # create a rule for compiling each file from eci since they cannot use
            # precompiled headers :(
            no_precompile = []
            for f in list(no_precompile_cfiles):
                f = m.pathrel(py.path.local(f))
                if f not in no_precompile and (f.endswith('.c') or f.endswith('.cpp')):
                    no_precompile.append(f)
                    target = f[:f.rfind('.')] + '.obj'
                    rules.append((target, f,
                        '$(CC) /nologo $(CFLAGS) $(CFLAGSEXTRA) '
                        '/Fo%s /c %s $(INCLUDEDIRS)' %(target, f)))

        else:
            rules.append(('.c.obj', '',
                          '$(CC) /nologo $(CFLAGS) $(CFLAGSEXTRA) '
                          '/Fo$@ /c $< $(INCLUDEDIRS)'))
            if shared:
                rules.append(('main.obj', 'main.c',
                              '$(CC) /nologo $(CFLAGS) $(CFLAGSEXTRA) '
                              '/Fo$@ /c main.c $(INCLUDEDIRS)'))
                rules.append(('wmain.obj', 'wmain.c',
                              '$(CC) /nologo $(CFLAGS) $(CFLAGSEXTRA) '
                              '/Fo$@ /c wmain.c $(INCLUDEDIRS)'))

        icon = config.translation.icon if config else None
        manifest = config.translation.manifest if config else None
        if icon:
            shutil.copyfile(icon, str(path.join('icon.ico')))
            rc_file = path.join('icon.rc')
            rc_file.write('IDI_ICON1 ICON DISCARDABLE "icon.ico"')
            rules.append(('icon.res', 'icon.rc', 'rc icon.rc'))
        if manifest:
            shutil.copyfile(manifest, str(path.join('pypy.manifest')))


        for args in definitions:
            m.definition(*args)

        for rule in rules:
            m.rule(*rule)

        if len(headers_to_precompile) > 0:
            # at least from VS2013 onwards we need to include PCH
            # objects in the final link command
            linkobjs = 'stdafx.obj '
        else:
            linkobjs = ''
        if len(' '.join(rel_ofiles)) > 2048:
            # command line is limited in length, use a response file
            linkobjs += '@<<\n$(OBJECTS)\n<<'
        else:
            linkobjs += '$(OBJECTS)'
        extra_deps = []
        if icon and not shared:
            extra_deps.append('icon.res')
            linkobjs = 'icon.res ' + linkobjs
        if manifest and not shared:
            linkflags.append('/MANIFESTINPUT:pypy.manifest')
        m.rule('$(TARGET)', ['$(OBJECTS)'] + extra_deps,
                [ '$(CC_LINK) $(LDFLAGS) $(LDFLAGSEXTRA)' +
                  ' $(LINKFILES) /out:$@ $(LIBDIRS) $(LIBS) ' +
                  linkobjs,
                ])
        m.rule('debugmode_$(TARGET)', ['$(OBJECTS)'] + extra_deps,
                [ '$(CC_LINK) /DEBUG $(LDFLAGS) $(LDFLAGSEXTRA)' +
                  ' $(LINKFILES) /out:$@ $(LIBDIRS) $(LIBS) ' +
                  linkobjs,
                ])

        if shared:
            m.definition('SHARED_IMPORT_LIB', so_name.new(ext='lib').basename)
            m.definition('PYPY_MAIN_FUNCTION', "pypy_main_startup")
            m.rule('main.c', '',
                   'echo '
                   'typedef unsigned short ARGV_T; '
                   'int $(PYPY_MAIN_FUNCTION)(int, ARGV_T*[]); '
                   'int wmain(int argc, ARGV_T* argv[]) '
                   '{ return $(PYPY_MAIN_FUNCTION)(argc, argv); } > $@')
            deps = ['main.obj']
            m.rule('wmain.c', '',
                   ['echo #define WIN32_LEAN_AND_MEAN > $@.tmp',
                   'echo #include "stdlib.h" >> $@.tmp',
                   'echo #include "windows.h" >> $@.tmp',
                   'echo int $(PYPY_MAIN_FUNCTION)(int, wchar_t*[]); >> $@.tmp',
                   'echo int WINAPI wWinMain( >> $@.tmp',
                   'echo     HINSTANCE hInstance,      /* handle to current instance */ >> $@.tmp',
                   'echo     HINSTANCE hPrevInstance,  /* handle to previous instance */ >> $@.tmp',
                   'echo     LPWSTR lpCmdLine,         /* pointer to command line */ >> $@.tmp',
                   'echo     int nCmdShow              /* show state of window */ >> $@.tmp',
                   'echo ) >> $@.tmp',
                   'echo { return $(PYPY_MAIN_FUNCTION)(__argc, __wargv); } >> $@.tmp',
                   'move $@.tmp $@',
                   ])
            wdeps = ['wmain.obj']
            if icon:
                deps.append('icon.res')
                wdeps.append('icon.res')
            manifest_args = '/MANIFEST:EMBED'
            if manifest:
                manifest_args += ' /MANIFESTINPUT:pypy.manifest'
            m.rule('$(DEFAULT_TARGET)', ['$(TARGET)'] + deps,
                   ['$(CC_LINK) /DEBUG /LARGEADDRESSAWARE /STACK:3145728 ' +
                    ' '.join(deps) + ' $(SHARED_IMPORT_LIB) ' +
                    manifest_args + ' /out:$@ '
                    ])
            m.rule('$(WTARGET)', ['$(TARGET)'] + wdeps,
                   ['$(CC_LINK) /DEBUG /LARGEADDRESSAWARE /STACK:3145728 ' +
                    '/SUBSYSTEM:WINDOWS '  +
                    ' '.join(wdeps) + ' $(SHARED_IMPORT_LIB) ' +
                    manifest_args + ' /out:$@ '
                    ])
            m.rule('debugmode_$(DEFAULT_TARGET)', ['debugmode_$(TARGET)']+deps,
                   ['$(CC_LINK) /DEBUG /LARGEADDRESSAWARE /STACK:3145728 ' +
                    ' '.join(deps) + ' debugmode_$(SHARED_IMPORT_LIB) ' +
                    manifest_args + ' /out:$@ '
                    ])

        return m

    def execute_makefile(self, path_to_makefile, extra_opts=[]):
        if isinstance(path_to_makefile, NMakefile):
            path = path_to_makefile.makefile_dir
        else:
            path = path_to_makefile
        log.execute('%s %s in %s' % (self.make, " ".join(extra_opts), path))
        oldcwd = path.chdir()
        try:
            returncode, stdout, stderr = _run_subprocess(
                self.make,
                ['/nologo', '/f', str(path.join('Makefile'))] + extra_opts,
                env = self.c_environ)
        finally:
            oldcwd.chdir()

        self._handle_error(returncode, stdout, stderr, path.join('make'))

# These are the external libraries, created and maintained by get_externals.py
# The buildbot runs get_externals before building
def patch_os_env(externals = Platform.externals):
    #print 'adding %s to PATH, INCLUDE, LIB' % basepath
    binpath = externals + r'\bin'
    path = os.environ['PATH']
    if binpath not in path:
        path = binpath + ';' + path
        # make sure externals is in current path for tests and translating
        os.environ['PATH'] = path
    if externals not in os.environ.get('INCLUDE', ''):
        os.environ['INCLUDE'] = externals + r'\include;' + os.environ.get('INCLUDE', '')
    if externals not in os.environ.get('LIB', ''):
        os.environ['LIB'] = externals + r'\lib;' + os.environ.get('LIB', '')
    return None

class WinDefinition(posix.Definition):
    def write(self, f):
        def write_list(prefix, lst):
            lst = lst or ['']
            for i, fn in enumerate(lst):
                print(prefix, fn, end=" ", file=f)
                if i < len(lst)-1:
                    print('\\', file=f)
                else:
                    print(file=f)
                prefix = ' ' * len(prefix)
        name, value = self.name, self.value
        if isinstance(value, str):
            f.write('%s = %s\n' % (name, value))
        else:
            write_list('%s =' % (name,), value)
        f.write('\n')


class NMakefile(posix.GnuMakefile):
    def write(self, out=None):
        # nmake expands macros when it parses rules.
        # Write all macros before the rules.
        if out is None:
            f = self.makefile_dir.join('Makefile').open('w')
        else:
            f = out
        for line in self.lines:
            if not isinstance(line, posix.Rule):
                line.write(f)
        for line in self.lines:
            if isinstance(line, posix.Rule):
                line.write(f)
        f.flush()
        if out is None:
            f.close()

    def definition(self, name, value):
        defs = self.defs
        defn = WinDefinition(name, value)
        if name in defs:
            self.lines[defs[name]] = defn
        else:
            defs[name] = len(self.lines)
            self.lines.append(defn)

class MingwPlatform(posix.BasePosix):
    name = 'mingw32'
    standalone_only = ()
    shared_only = ()
    cflags = ('-O3',)
    link_flags = ()
    exe_ext = 'exe'
    so_ext = 'dll'

    def __init__(self, cc=None):
        if not cc:
            cc = 'gcc'
        Platform.__init__(self, cc)

    def _args_for_shared(self, args, **kwds):
        return ['-shared'] + args

    def _include_dirs_for_libffi(self):
        return []

    def _library_dirs_for_libffi(self):
        return []

    def _handle_error(self, returncode, stdout, stderr, outname):
        # Mingw tools write compilation errors to stdout
        super(MingwPlatform, self)._handle_error(
            returncode, '', stderr + stdout, outname)
